from builtins import object
import os
from pynamodb.models import Model
from pynamodb.attributes import UnicodeAttribute
from pynamodb.indexes import GlobalSecondaryIndex, AllProjection

"""
    DbProperty handles all db operations for a property
    AWS DynamoDB is used as a backend.
"""


class PropertyIndex(GlobalSecondaryIndex):
    """
    Secondary index on property
    """
    class Meta(object):
        index_name = 'property-index'
        read_capacity_units = 2
        write_capacity_units = 1
        projection = AllProjection()

    value = UnicodeAttribute(default=0, hash_key=True)


class Property(Model):
    """
       DynamoDB data model for a property
    """
    class Meta(object):
        table_name = os.getenv('AWS_DB_PREFIX', 'demo_actingweb') + "_properties"
        read_capacity_units = 26
        write_capacity_units = 2
        region = os.getenv('AWS_DEFAULT_REGION', 'us-west-1')
        host = os.getenv('AWS_DB_HOST', None)

    id = UnicodeAttribute(hash_key=True)
    name = UnicodeAttribute(range_key=True)
    value = UnicodeAttribute()
    property_index = PropertyIndex()


class DbProperty(object):
    """
        DbProperty does all the db operations for property objects

        The actor_id must always be set. get(), set() and
        get_actor_id_from_property() will set a new internal handle
        that will be reused by set() (overwrite property) and
        delete().
    """

    def get(self,  actor_id=None, name=None):
        """ Retrieves the property from the database """
        if not actor_id or not name:
            return None
        if self.handle:
            try:
                self.handle.refresh()
            except Property.DoesNotExist:
                return None
            return self.handle.value
        try:
            self.handle = Property.get(actor_id, name, consistent_read=True)
        except Property.DoesNotExist:
            return None
        return self.handle.value

    def get_actor_id_from_property(self, name=None, value=None):
        """ Retrives an actor_id based on the value of a property.
        """
        if not name or not value:
            return None
        results = Property.property_index.query(value)
        self.handle = None
        for res in results:
            self.handle = res
            break
        if not self.handle:
            return None
        return self.handle.id

    def set(self, actor_id=None, name=None, value=None):
        """ Sets a new value for the property name
        """
        if not name:
            return False
        if not value or len(value) == 0:
            if self.get(actor_id=actor_id, name=name):
                self.delete()
            return True
        if not self.handle:
            if not actor_id:
                return False
            self.handle = Property(id=actor_id, name=name,
                                   value=value)
        else:
            self.handle.value = value
        self.handle.save()
        return True

    def delete(self):
        """ Deletes the property in the database after a get() """
        if not self.handle:
            return False
        self.handle.delete()
        self.handle = None
        return True

    def __init__(self):
        self.handle = None
        if not Property.exists():
            Property.create_table(wait=True)


class DbPropertyList(object):
    """
        DbPropertyList does all the db operations for list of property objects

        The actor_id must always be set.
    """

    def fetch(self,  actor_id=None):
        """ Retrieves the properties of an actor_id from the database """
        if not actor_id:
            return None
        self.actor_id = actor_id
        self.handle = Property.scan(Property.id == actor_id)
        if self.handle:
            self.props = {}
            for d in self.handle:
                self.props[d.name] = d.value
            return self.props
        else:
            return None

    def delete(self):
        """ Deletes all the properties in the database """
        if not self.actor_id:
            return False
        self.handle = Property.scan(Property.id == self.actor_id)
        if not self.handle:
            return False
        for p in self.handle:
            p.delete()
        self.handle = None
        return True

    def __init__(self):
        self.handle = None
        self.actor_id = None
        self.props = None



from builtins import str
from builtins import object
import logging
import datetime
import os

from pynamodb.models import Model
from pynamodb.attributes import UnicodeAttribute, NumberAttribute, UTCDateTimeAttribute

"""
    DbSubscriptionDiff handles all db operations for a subscription diff

    DbSubscriptionDiffList handles list of subscriptions diffs
    Google datastore for google is used as a backend.
"""


class SubscriptionDiff(Model):
    class Meta(object):
        table_name = os.getenv('AWS_DB_PREFIX', 'demo_actingweb') + "_subscriptiondiffs"
        read_capacity_units = 2
        write_capacity_units = 3
        region = os.getenv('AWS_DEFAULT_REGION', 'us-west-1')
        host = os.getenv('AWS_DB_HOST', None)

    id = UnicodeAttribute(hash_key=True)
    subid_seqnr = UnicodeAttribute(range_key=True)
    subid = UnicodeAttribute()
    timestamp = UTCDateTimeAttribute(default=datetime.datetime.utcnow())
    diff = UnicodeAttribute()
    seqnr = NumberAttribute(default=1)


class DbSubscriptionDiff(object):
    """
        DbSubscriptionDiff does all the db operations for subscription diff objects

        The  actor_id must always be set.
    """

    def get(self,  actor_id=None, subid=None, seqnr=None):
        """ Retrieves the subscriptiondiff from the database """
        if not actor_id and not self.handle:
            return None
        if not subid and not self.handle:
            logging.debug("Attempt to get subscriptiondiff without subid")
            return None
        if not self.handle:
            if not seqnr:
                query = SubscriptionDiff.query(
                    actor_id,
                    SubscriptionDiff.subid_seqnr.startswith(subid),
                    consistent_read=True)
                # Find the record with lowest seqnr
                for t in query:
                    if not self.handle:
                        self.handle = t
                        continue
                    if t.seqnr < self.handle.seqnr:
                        self.handle = t
            else:
                self.handle = SubscriptionDiff.get(
                    actor_id,
                    subid + ":" + str(seqnr),
                    consistent_read=True)
        if self.handle:
            t = self.handle
            return {
                "id": t.id,
                "subscriptionid": t.subid,
                "timestamp": t.timestamp,
                "data": t.diff,
                "sequence": t.seqnr,
            }
        else:
            return None

    def create(self, actor_id=None,
               subid=None,
               diff='',
               seqnr=1):
        """ Create a new subscription diff """
        if not actor_id or not subid:
            logging.debug("Attempt to create subscriptiondiff without actorid or subid")
            return False
        self.handle = SubscriptionDiff(id=actor_id,
                                       subid_seqnr=subid + ":" + str(seqnr),
                                       subid=subid,
                                       diff=diff,
                                       seqnr=seqnr)
        self.handle.save()
        return True

    def delete(self):
        """ Deletes the subscription diff in the database """
        if not self.handle:
            return False
        self.handle.delete()
        self.handle = None
        return True

    def __init__(self):
        self.handle = None
        if not SubscriptionDiff.exists():
            SubscriptionDiff.create_table(wait=True)


class DbSubscriptionDiffList(object):
    """
        DbSubscriptionDiffList does all the db operations for list of diff objects

        The actor_id must always be set. 
    """

    def fetch(self, actor_id=None, subid=None):
        """ Retrieves the subscription diffs of an actor_id from the database as an array"""
        if not actor_id:
            return None
        self.actor_id = actor_id
        self.subid = subid
        if not subid:
            self.handle = SubscriptionDiff.query(
                actor_id,
                consistent_read=True)
        else:
            self.handle = SubscriptionDiff.query(
                actor_id,
                SubscriptionDiff.subid.startswith(subid),
                consistent_read=True)
        self.diffs = []
        if self.handle:
            for t in self.handle:
                self.diffs.append({
                    "id": t.id,
                    "subscriptionid": t.subid,
                    "timestamp": t.timestamp,
                    "diff": t.diff,
                    "sequence": t.seqnr,
                })
                sorted(self.diffs, key=lambda diff: diff["sequence"])
            return self.diffs
        else:
            return []

    def delete(self, seqnr=None):
        """ Deletes all the fetched subscription diffs in the database 

            Optional seqnr deletes up to (excluding) a specific seqnr
        """
        if not self.handle:
            return False
        if not seqnr or not isinstance(seqnr, int):
            seqnr = 0
        if not self.subid:
            self.handle = SubscriptionDiff.query(
                self.actor_id,
                consistent_read=True)
        else:
            self.handle = SubscriptionDiff.query(
                self.actor_id,
                SubscriptionDiff.subid.startswith(self.subid),
                consistent_read=True)
        for p in self.handle:
            if seqnr == 0 or p.seqnr <= seqnr:
                p.delete()
        self.handle = None
        return True

    def __init__(self):
        self.handle = None
        self.diffs = []
        self.actor_id = None
        self.subid = None
        if not SubscriptionDiff.exists():
            SubscriptionDiff.create_table(wait=True)



from builtins import object
import logging
import os
from pynamodb.models import Model
from pynamodb.attributes import UnicodeAttribute, NumberAttribute, BooleanAttribute

"""
    DbSubscription handles all db operations for a subscription

    DbSubscriptionList handles list of subscriptions
    AWS Dynamodb is used as a backend.
"""


class Subscription(Model):
    class Meta(object):
        table_name = os.getenv('AWS_DB_PREFIX', 'demo_actingweb') + "_subscriptions"
        read_capacity_units = 2
        write_capacity_units = 1
        region = os.getenv('AWS_DEFAULT_REGION', 'us-west-1')
        host = os.getenv('AWS_DB_HOST', None)

    id = UnicodeAttribute(hash_key=True)
    peer_sub_id = UnicodeAttribute(range_key=True)
    peerid = UnicodeAttribute()
    subid = UnicodeAttribute()
    granularity = UnicodeAttribute(null=True)
    target = UnicodeAttribute(null=True)
    subtarget = UnicodeAttribute(null=True)
    resource = UnicodeAttribute(null=True)
    seqnr = NumberAttribute(default=1)
    callback = BooleanAttribute()


class DbSubscription(object):
    """
        DbSubscription does all the db operations for subscription objects

        The  actor_id must always be set.
    """

    def get(self,  actor_id=None, peerid=None, subid=None):
        """ Retrieves the subscription from the database """
        if not actor_id:
            return None
        if not peerid or not subid:
            logging.debug("Attempt to get subscription without peerid or subid")
            return None
        try:
            # We only expect one
            for t in Subscription.query(actor_id,
                                        Subscription.peer_sub_id == peerid + ":" + subid,
                                        consistent_read=True):
                self.handle = t
                return {
                    "id": t.id,
                    "peerid": t.peerid,
                    "subscriptionid": t.subid,
                    "granularity": (t.granularity or ''),
                    "target": (t.target or ''),
                    "subtarget": (t.subtarget or ''),
                    "resource": (t.resource or ''),
                    "sequence": t.seqnr,
                    "callback": t.callback,
                }
        except Subscription.DoesNotExist:
            pass
        return None

    def modify(self,
               peerid=None,
               subid=None,
               granularity=None,
               target=None,
               subtarget=None,
               resource=None,
               seqnr=None,
               callback=None):
        """ Modify a subscription
            If bools are none, they will not be changed.
        """
        if not self.handle:
            logging.debug("Attempted modification of DbSubscription without db handle")
            return False
        if peerid and len(peerid) > 0:
            self.handle.peerid = peerid
        if subid and len(subid) > 0:
            self.handle.subid = subid
        if granularity and len(granularity) > 0:
            self.handle.granularity = granularity
        if callback is not None:
            self.handle.callback = callback
        if target and len(target) > 0:
            self.handle.target = target
        if subtarget and len(subtarget) > 0:
            self.handle.subtarget = subtarget
        if resource and len(resource) > 0:
            self.handle.resource = resource
        if seqnr:
            self.handle.seqnr = seqnr
        self.handle.save()
        return True

    def create(self, actor_id=None,
               peerid=None,
               subid=None,
               granularity=None,
               target=None,
               subtarget=None,
               resource=None,
               seqnr=1,
               callback=False):
        """ Create a new subscription """
        if not actor_id or not peerid or not subid:
            return False
        if self.get(actor_id=actor_id, peerid=peerid, subid=subid):
            return False
        self.handle = Subscription(id=actor_id,
                                   peer_sub_id=peerid + ":" + subid,
                                   peerid=peerid,
                                   subid=subid,
                                   seqnr=seqnr,
                                   callback=callback)
        if granularity and len(granularity) > 0:
            self.handle.granularity = granularity
        if target and len(target) > 0:
            self.handle.target = target
        if subtarget and len(subtarget) > 0:
            self.handle.subtarget = subtarget
        if resource and len(resource) > 0:
            self.handle.resource = resource
        self.handle.save()
        return True

    def delete(self):
        """ Deletes the subscription in the database """
        if not self.handle:
            logging.debug("Attempted delete of DbSubscription with no handle set.")
            return False
        self.handle.delete()
        self.handle = None
        return True

    def __init__(self):
        self.handle = None
        if not Subscription.exists():
            Subscription.create_table(wait=True)


class DbSubscriptionList(object):
    """
        DbTrustList does all the db operations for list of trust objects

        The  actor_id must always be set.
    """

    def fetch(self, actor_id):
        """ Retrieves the subscriptions of an actor_id from the database as an array"""
        if not actor_id:
            return None
        self.actor_id = actor_id
        self.handle = Subscription.query(self.actor_id, consistent_read=True)
        self.subscriptions = []
        if self.handle:
            for t in self.handle:
                self.subscriptions.append({
                    "id": t.id,
                    "peerid": t.peerid,
                    "subscriptionid": t.subid,
                    "granularity": (t.granularity or ''),
                    "target": (t.target or ''),
                    "subtarget": (t.subtarget or ''),
                    "resource": (t.resource or ''),
                    "sequence": t.seqnr,
                    "callback": t.callback,
                })
            return self.subscriptions
        else:
            return []

    def delete(self):
        """ Deletes all the subscriptions for an actor in the database """
        if not self.actor_id:
            return False
        self.handle = Subscription.query(self.actor_id, consistent_read=True)
        if not self.handle:
            return False
        for p in self.handle:
            p.delete()
        self.handle = None
        return True

    def __init__(self):
        self.handle = None
        self.actor_id = None
        self.subscriptions = []
        if not Subscription.exists():
            Subscription.create_table(wait=True)



from builtins import object
import os
from pynamodb.models import Model
from pynamodb.attributes import UnicodeAttribute, JSONAttribute, UTCDateTimeAttribute

"""
    DbAttribute handles all db operations for an attribute (internal)
    AWS DynamoDB is used as a backend.
"""


class Attribute(Model):
    """
       DynamoDB data model for a property
    """
    class Meta(object):
        table_name = os.getenv('AWS_DB_PREFIX', 'demo_actingweb') + "_attributes"
        read_capacity_units = 26
        write_capacity_units = 2
        region = os.getenv('AWS_DEFAULT_REGION', 'us-west-1')
        host = os.getenv('AWS_DB_HOST', None)

    id = UnicodeAttribute(hash_key=True)
    bucket_name = UnicodeAttribute(range_key=True)
    bucket = UnicodeAttribute()
    name = UnicodeAttribute()
    data = JSONAttribute(null=True)
    timestamp = UTCDateTimeAttribute(null=True)


class DbAttribute(object):
    """
        DbProperty does all the db operations for property objects

        The actor_id must always be set. get(), set() will set a new internal handle
        that will be reused by set() (overwrite attribute) and
        delete().
    """

    @staticmethod
    def get_bucket(actor_id=None, bucket=None):
        """ Returns a dict of attributes from a bucket, each with data and timestamp """
        if not actor_id or not bucket:
            return None
        try:
            query = Attribute.query(
                actor_id,
                Attribute.bucket_name.startswith(bucket),
                consistent_read=True)
        except Attribute.DoesNotExist:
            return None
        ret = {}
        for t in query:
            ret[t.name] = {
                "data": t.data,
                "timestamp": t.timestamp,
            }
        return ret

    @staticmethod
    def get_attr(actor_id=None, bucket=None, name=None):
        """ Returns a dict of attributes from a bucket, each with data and timestamp """
        if not actor_id or not bucket or not name:
            return None
        try:
            r = Attribute.get(actor_id, bucket + ":" + name, consistent_read=True)
        except Attribute.DoesNotExist:
            return None
        return {
            "data": r.data,
            "timestamp": r.timestamp,
        }

    @staticmethod
    def set_attr(actor_id=None, bucket=None, name=None, data=None, timestamp=None):
        """ Sets a data value for a given attribute in a bucket
        """
        if not actor_id or not name or not bucket:
            return False
        if not data:
            try:
                item = Attribute.get(actor_id, bucket + ":" + name, consistent_read=True)
                item.delete()
            except Attribute.DoesNotExist:
                pass
            return True
        new = Attribute(
            id=actor_id,
            bucket_name=bucket + ":" + name,
            bucket=bucket,
            name=name,
            data=data,
            timestamp=timestamp
        )
        new.save()
        return True

    def delete_attr(self, actor_id=None, bucket=None, name=None):
        """ Deletes an attribute in a bucket
        """
        return self.set_attr(actor_id=actor_id, bucket=bucket, name=name, data=None)

    @staticmethod
    def delete_bucket(actor_id=None, bucket=None):
        """ Deletes an entire bucket
        """
        if not actor_id or not bucket:
            return False
        try:
            query = Attribute.query(
                actor_id,
                Attribute.bucket_name.startswith(bucket),
                consistent_read=True)
        except Attribute.DoesNotExist:
            return True
        for t in query:
            t.delete()
        return True

    def __init__(self):
        if not Attribute.exists():
            Attribute.create_table(wait=True)


class DbAttributeBucketList(object):
    """
        DbAttributeBucketList handles multiple buckets

        The  actor_id must always be set.
    """

    @staticmethod
    def fetch(actor_id=None):
        """ Retrieves all the attributes of an actor_id from the database """
        if not actor_id:
            return None
        try:
            query = Attribute.query(actor_id)
        except Attribute.DoesNotExist:
            return None
        ret = {}
        for t in query:
            if t.bucket not in ret:
                ret[t.bucket] = {}
            ret[t.bucket][t.name] = {
                "data": t.data,
                "timestamp": t.timestamp,
            }
        return ret

    @staticmethod
    def delete(actor_id=None):
        """ Deletes all the attributes in the database """
        if not actor_id:
            return False
        try:
            query = Attribute.query(actor_id)
        except Attribute.DoesNotExist:
            return False
        for t in query:
            t.delete()
        return True

    def __init__(self):
        if not Attribute.exists():
            Attribute.create_table(wait=True)



from __future__ import absolute_import
from builtins import str
from builtins import object
import datetime
import base64
import logging
import json
from actingweb import (property, trust, subscription, peertrustee, attribute)


class DummyPropertyClass:
    """ Only used to deprecate get_property() in 2.4.4 """

    def __init__(self, v=None):
        self.value = v


class Actor(object):

    ###################
    # Basic operations
    ###################

    def __init__(self, actor_id=None, config=None):
        self.config = config
        self.property_list = None
        self.subs_list = None
        self.actor = None
        self.passphrase = None
        self.creator = None
        self.last_response_code = 0
        self.last_response_message = ''
        self.id = actor_id
        self.handle = self.config.DbActor.DbActor()
        if actor_id and config:
            self.store = attribute.InternalStore(actor_id=actor_id, config=config)
            self.property = property.PropertyStore(actor_id=actor_id, config=config)
        else:
            self.store = None
            self.property = None
        self.get(actor_id=actor_id)

    def get_peer_info(self, url: str) -> dict:
        """ Contacts an another actor over http/s to retrieve meta information
        :param url: Root URI of a remote actor
        :rtype: dict
        :return: The json response from the /meta path in the data element and last_response_code/last_response_message
        set to the results of the https request
        :Example:

        >>>{
        >>>    "last_response_code": 200,
        >>>    "last_response_message": "OK",
        >>>    "data":{}
        >>>}
        """
        try:
            logging.debug('Getting peer info at url(' + url + ')')
            if self.config.env == "appengine":
                self.config.module["urlfetch"].set_default_fetch_deadline(20)
            response = self.config.module["urlfetch"].fetch(url=url + '/meta')
            res = {
                "last_response_code": response.status_code,
                "last_response_message": response.content,
                "data": json.loads(response.content.decode('utf-8', 'ignore')),
            }
            logging.debug('Got peer info from url(' + url +
                          ') with body(' + str(response.content) + ')')
        except (TypeError, ValueError, KeyError):
            res = {
                "last_response_code": 500,
            }
        return res

    def get(self, actor_id: str = None) -> dict or None:
        """Retrieves an actor from storage or initialises if it does not exist"""
        if not actor_id and not self.id:
            return None
        elif not actor_id:
            actor_id = self.id
        if self.handle and self.actor and len(self.actor) > 0:
            return self.actor
        self.actor = self.handle.get(actor_id=actor_id)
        if self.actor and len(self.actor) > 0:
            self.id = self.actor["id"]
            self.creator = self.actor["creator"]
            self.passphrase = self.actor["passphrase"]
            self.store = attribute.InternalStore(actor_id=self.id, config=self.config)
            self.property = property.PropertyStore(actor_id=self.id, config=self.config)
            if self.config.force_email_prop_as_creator:
                em = self.store.email
                if self.config.migrate_2_5_0 and not em:
                    em = self.property.email
                    if em:
                        self.store.email = em.lower()
                        self.property.email = None
                if em and em.lower() != self.creator:
                    self.modify(creator=em.lower())
        else:
            self.id = None
            self.creator = None
            self.passphrase = None
        return self.actor

    def get_from_property(self, name='oauthId', value=None):
        """ Initialise an actor by matching on a stored property.

        Use with caution as the property's value de-facto becomes
        a security token. If multiple properties are found with the
        same value, no actor will be initialised.
        Also note that this is a costly operation as all properties
        of this type will be retrieved and proceessed.
        """
        actor_id = property.Property(name=name, value=value, config=self.config).get_actor_id()
        if not actor_id:
            self.id = None
            self.creator = None
            self.passphrase = None
            return
        self.get(actor_id=actor_id)

    def get_from_creator(self, creator=None):
        """ Initialise an actor by matching on creator.

        If unique_creator config is False, then no actor will be initialised.
        Likewise, if multiple properties are found with the same value (due to earlier
        uniqueness off).
        """
        self.id = None
        self.creator = None
        self.passphrase = None
        if not self.config.unique_creator:
            return False
        exists = self.config.DbActor.DbActor().get_by_creator(creator=creator)
        if len(exists) != 1:
            return False
        self.get(actor_id=exists[0]["id"])
        return True

    def create(self, url, creator, passphrase, actor_id=None, delete=False):
        """"Creates a new actor and persists it.

            If delete is True, any existing actors with same creator value
            will be deleted. If it is False, the one with the correct passphrase
            will be chosen (if any)
        """
        seed = url
        now = datetime.datetime.utcnow()
        seed += now.strftime("%Y%m%dT%H%M%S%f")
        if len(creator) > 0:
            self.creator = creator
        else:
            self.creator = "creator"
        if self.config.unique_creator:
            in_db = self.config.DbActor.DbActor()
            exists = in_db.get_by_creator(creator=self.creator)
            if exists:
                # If uniqueness is turned on at a later point, we may have multiple accounts
                # with creator as "creator". Check if we have an internal value "email" and then
                # set creator to the email address.
                if delete:
                    for c in exists:
                        anactor = Actor(actor_id=c["id"], config=self.config)
                        anactor.delete()
                else:
                    if self.config.force_email_prop_as_creator and self.creator == "creator":
                        for c in exists:
                            anactor = Actor(actor_id=c["id"])
                            em = anactor.store.email
                            if self.config.migrate_2_5_0 and not em:
                                em = anactor.property.email
                                if em:
                                    anactor.store.email = em.lower()
                                    anactor.property.email = None
                            if em:
                                anactor.modify(creator=em.lower())
                    for c in exists:
                        if c['passphrase'] == passphrase:
                            self.handle = in_db
                            self.id = c['id']
                            self.passphrase = c['passphrase']
                            self.creator = c['creator']
                            return True
                        return False
        if passphrase and len(passphrase) > 0:
            self.passphrase = passphrase
        else:
            self.passphrase = self.config.new_token()
        if actor_id:
            self.id = actor_id
        else:
            self.id = self.config.new_uuid(seed)
        if not self.handle:
            self.handle = self.config.DbActor.DbActor()
        self.handle.create(creator=self.creator,
                           passphrase=self.passphrase,
                           actor_id=self.id)
        self.store = attribute.InternalStore(actor_id=self.id, config=self.config)
        self.property = property.PropertyStore(actor_id=self.id, config=self.config)
        return True

    def modify(self, creator=None):
        if not self.handle or not creator:
            logging.debug("Attempted modify of actor with no handle or no param changed")
            return False
        if '@' in creator:
            creator = creator.lower()
        self.creator = creator
        if self.actor:
            self.actor["creator"] = creator
        self.handle.modify(creator=creator)
        return True

    def delete(self):
        """Deletes an actor and cleans up all relevant stored data"""
        if not self.handle:
            logging.debug("Attempted delete of actor with no handle")
            return False
        self.delete_peer_trustee(shorttype='*')
        if not self.property_list:
            self.property_list = property.Properties(actor_id=self.id, config=self.config)
        self.property_list.delete()
        subs = subscription.Subscriptions(actor_id=self.id, config=self.config)
        subs.fetch()
        subs.delete()
        trusts = trust.Trusts(actor_id=self.id, config=self.config)
        relationships = trusts.fetch()
        for rel in relationships:
            self.delete_reciprocal_trust(peerid=rel["peerid"], delete_peer=True)
        trusts.delete()
        buckets = attribute.Buckets(actor_id=self.id, config=self.config)
        buckets.delete()
        self.handle.delete()

    ######################
    # Advanced operations
    ######################

    def set_property(self, name, value):
        """Sets an actor's property name to value. (DEPRECATED, use actor's property store!) """
        self.property[name] = value

    def get_property(self, name):
        """Retrieves a property object named name. (DEPRECATED, use actor's property store!) """
        return DummyPropertyClass(self.property[name])

    def delete_property(self, name):
        """Deletes a property name. (DEPRECATED, use actor's property store!)"""
        self.property[name] = None

    def delete_properties(self):
        """Deletes all properties."""
        if not self.property_list:
            self.property_list = property.Properties(actor_id=self.id, config=self.config)
        return self.property_list.delete()

    def get_properties(self):
        """Retrieves properties from db and returns a dict."""
        self.property_list = property.Properties(actor_id=self.id, config=self.config)
        return self.property_list.fetch()

    def delete_peer_trustee(self, shorttype=None, peerid=None):
        if not peerid and not shorttype:
            return False
        if shorttype == '*':
            for t in self.config.actors:
                self.delete_peer_trustee(shorttype=t)
            return True
        if shorttype and not self.config.actors[shorttype]:
            logging.error('Got a request to delete an unknown actor type(' + shorttype + ')')
            return False
        peer_data = None
        new_peer = None
        if peerid:
            new_peer = peertrustee.PeerTrustee(actor_id=self.id, peerid=peerid, config=self.config)
            peer_data = new_peer.get()
            if not peer_data or len(peer_data) == 0:
                return False
        elif shorttype:
            new_peer = peertrustee.PeerTrustee(actor_id=self.id, short_type=shorttype, config=self.config)
            peer_data = new_peer.get()
            if not peer_data or len(peer_data) == 0:
                return False
        logging.debug(
            'Deleting peer actor at baseuri(' + peer_data["baseuri"] + ')')
        u_p = b'trustee:' + peer_data["passphrase"].encode('utf-8')
        headers = {'Authorization': 'Basic ' +
                   base64.b64encode(u_p).decode('utf-8'),
                   }
        try:
            if self.config.env == 'appengine':
                self.config.module["urlfetch"].set_default_fetch_deadline(20)
                response = self.config.module["urlfetch"].fetch(url=peer_data["baseuri"],
                                                                method=self.config.module["urlfetch"].DELETE,
                                                                headers=headers
                                                                )
            else:
                response = self.config.module["urlfetch"].delete(url=peer_data["baseuri"], headers=headers)
            self.last_response_code = response.status_code
            self.last_response_message = response.content
        except (self.config.module["urlfetch"].UrlfetchException,
                self.config.module["urlfetch"].URLError,
                self.config.module["urlfetch"].Timeout,
                self.config.module["urlfetch"].TooManyRedirects):
            logging.debug('Not able to delete peer actor remotely')
            self.last_response_code = 408
            return False
        if response.status_code < 200 or response.status_code > 299:
            logging.debug('Not able to delete peer actor remotely')
            return False
        # Delete trust, peer is already deleted remotely
        if not self.delete_reciprocal_trust(peerid=peer_data["peerid"], delete_peer=False):
            logging.debug('Not able to delete peer actor trust in db')
        if not new_peer.delete():
            logging.debug('Not able to delete peer actor in db')
            return False
        return True

    def get_peer_trustee(self, shorttype=None, peerid=None):
        """ Get a peer, either existing or create it as trustee 

        Will retrieve an existing peer or create a new and establish trust.
        If no trust exists, a new trust will be established.
        Use either peerid to target a specific known peer, or shorttype to
        allow creation of a new peer if none exists
        """
        if not peerid and not shorttype:
            return None
        if shorttype and not self.config.actors[shorttype]:
            logging.error('Got a request to create an unknown actor type(' + shorttype + ')')
            return None
        if peerid:
            new_peer = peertrustee.PeerTrustee(actor_id=self.id, peerid=peerid, config=self.config)
        else:
            new_peer = peertrustee.PeerTrustee(actor_id=self.id, short_type=shorttype, config=self.config)
        peer_data = new_peer.get()
        if peer_data and len(peer_data) > 0:
            logging.debug('Found peer in getPeer, now checking existing trust...')
            dbtrust = trust.Trust(actor_id=self.id, peerid=peer_data["peerid"], config=self.config)
            new_trust = dbtrust.get()
            if new_trust and len(new_trust) > 0:
                return peer_data
            logging.debug('Did not find existing trust, will create a new one')
        factory = self.config.actors[shorttype]['factory']
        # If peer did not exist, create it as trustee
        if not peer_data or len(peer_data) == 0:
            if len(factory) == 0:
                logging.error('Peer actor of shorttype(' + 
                              shorttype + ') does not have factory set.')
            params = {
                'creator': 'trustee',
                'trustee_root': self.config.root + self.id
            }
            data = json.dumps(params)
            logging.debug(
                'Creating peer actor at factory(' + factory + ') with data(' +
                str(data) + ')')
            response = None
            try:
                if self.config.env == 'appengine':
                    self.config.module["urlfetch"].set_default_fetch_deadline(20)
                    response = self.config.module["urlfetch"].fetch(url=factory,
                                                                    method=self.config.module["urlfetch"].POST,
                                                                    payload=data
                                                                    )
                else:
                    response = self.config.module["urlfetch"].post(
                        url=factory,
                        data=data,
                        headers={
                            'Content-Type': 'application/json'
                        }
                    )
                self.last_response_code = response.status_code
                self.last_response_message = response.content
            except (self.config.module["urlfetch"].UrlfetchException,
                    self.config.module["urlfetch"].URLError,
                    self.config.module["urlfetch"].Timeout,
                    self.config.module["urlfetch"].TooManyRedirects):
                logging.debug('Not able to create new peer actor')
                self.last_response_code = 408
            logging.debug('Create peer actor POST response:' + str(self.last_response_code))
            if self.last_response_code < 200 or self.last_response_code > 299:
                return None
            try:
                data = json.loads(response.content.decode('utf-8', 'ignore'))
            except (TypeError, ValueError, KeyError):
                logging.warning("Not able to parse response when creating peer at factory(" +
                                factory + ")")
                return None
            if 'Location' in response.headers:
                baseuri = response.headers['Location']
            elif 'location' in response.headers:
                baseuri = response.headers['location']
            else:
                logging.warning("No location uri found in response when creating a peer as trustee")
                baseuri = ""
            res = self.get_peer_info(baseuri)
            if not res or res["last_response_code"] < 200 or res["last_response_code"] >= 300:
                return None
            info_peer = res['data']
            if not info_peer or ('id' in info_peer and not info_peer["id"])\
                    or ('type' in info_peer and not info_peer["type"]):
                logging.info(
                    'Received invalid peer info when trying to create peer actor at: ' + str(factory))
                return None
            new_peer = peertrustee.PeerTrustee(actor_id=self.id, peerid=info_peer["id"], peer_type=info_peer["type"],
                                               config=self.config)
            if not new_peer.create(baseuri=baseuri, passphrase=data["passphrase"]):
                logging.error('Failed to create in db new peer Actor(' +
                              self.id + ') at ' + baseuri)
                return None
        # Now peer exists, create trust
        new_peer_data = new_peer.get()
        new_trust = self.create_reciprocal_trust(
                        url=new_peer_data["baseuri"],
                        secret=self.config.new_token(),
                        desc='Trust from trustee to ' + shorttype,
                        relationship=self.config.actors[shorttype]['relationship']
                        )
        if not new_trust or len(new_trust) == 0:
            logging.warning("Not able to establish trust relationship with peer at factory(" +
                            factory + ")")
        else:
            # Approve the relationship
            params = {
                'approved': True,
            }
            u_p = b'trustee:' + new_peer_data["passphrase"].encode('utf-8')
            headers = {'Authorization': 'Basic ' +
                       base64.b64encode(u_p).decode('utf-8'),
                       'Content-Type': 'application/json',
                       }
            data = json.dumps(params)
            try:
                if self.config.env == 'appengine':
                    self.config.module["urlfetch"].set_default_fetch_deadline(20)
                    response = self.config.module["urlfetch"].fetch(url=new_peer_data["baseuri"] +
                                                                    '/trust/' +
                                                                    self.config.actors[shorttype]['relationship'] +
                                                                    '/' + self.id,
                                                                    method=self.config.module["urlfetch"].PUT,
                                                                    payload=data,
                                                                    headers=headers
                                                                    )
                else:
                    response = self.config.module["urlfetch"].put(
                        url=new_peer_data["baseuri"] + '/trust/' + self.config.actors[shorttype]['relationship'] +
                        '/' + self.id, data=data, headers=headers)
                self.last_response_code = response.status_code
                self.last_response_message = response.content
            except (self.config.module["urlfetch"].UrlfetchException,
                    self.config.module["urlfetch"].URLError,
                    self.config.module["urlfetch"].Timeout,
                    self.config.module["urlfetch"].TooManyRedirects):
                self.last_response_code = 408
                self.last_response_message = 'Not able to approve peer actor trust remotely'
            if self.last_response_code < 200 or self.last_response_code > 299:
                logging.debug('Not able to delete peer actor remotely')
        return new_peer_data

    def get_trust_relationship(self, peerid=None):
        if not peerid:
            return None
        return trust.Trust(actor_id=self.id, peerid=peerid, config=self.config).get()

    def get_trust_relationships(self, relationship='', peerid='', trust_type=''):
        """Retrieves all trust relationships or filtered."""
        trust_list = trust.Trusts(actor_id=self.id, config=self.config)
        relationships = trust_list.fetch()
        rels = []
        for rel in relationships:
            if len(relationship) > 0 and relationship != rel["relationship"]:
                continue
            if len(peerid) > 0 and peerid != rel["peerid"]:
                continue
            if len(trust_type) > 0 and trust_type != rel["type"]:
                continue
            rels.append(rel)
        return rels

    def modify_trust_and_notify(self, 
                                relationship=None, 
                                peerid=None, 
                                baseuri='', 
                                secret='', 
                                desc='', 
                                approved=None, 
                                verified=None, 
                                verification_token=None, 
                                peer_approved=None):
        """Changes a trust relationship and noties the peer if approval is changed."""
        if not relationship or not peerid:
            return False
        relationships = self.get_trust_relationships(
            relationship=relationship, peerid=peerid)
        if not relationships:
            return False
        this_trust = relationships[0]
        headers = dict()
        # If we change approval status, send the changed status to our peer
        if approved is True and this_trust["approved"] is False:
            params = {
                'approved': True,
            }
            requrl = this_trust["baseuri"] + '/trust/' + relationship + '/' + self.id
            if this_trust["secret"]:
                headers = {'Authorization': 'Bearer ' + this_trust["secret"],
                           'Content-Type': 'application/json',
                           }
            data = json.dumps(params)
            # Note the POST here instead of PUT. POST is used to used to notify about
            # state change in the relationship (i.e. not change the object as PUT
            # would do)
            logging.debug(
                'Trust relationship has been approved, notifying peer at url(' + requrl + ')')
            try:
                if self.config.env == 'appengine':
                    self.config.module["urlfetch"].set_default_fetch_deadline(20)
                    response = self.config.module["urlfetch"].fetch(url=requrl,
                                                                    method=self.config.module["urlfetch"].POST,
                                                                    payload=data,
                                                                    headers=headers
                                                                    )
                else:
                    response = self.config.module["urlfetch"].post(
                        url=requrl,
                        data=data,
                        headers=headers
                        )
                self.last_response_code = response.status_code
                self.last_response_message = response.content
            except (self.config.module["urlfetch"].UrlfetchException,
                    self.config.module["urlfetch"].URLError,
                    self.config.module["urlfetch"].Timeout,
                    self.config.module["urlfetch"].TooManyRedirects):
                logging.debug('Not able to notify peer at url(' + requrl + ')')
                self.last_response_code = 500
        dbtrust = trust.Trust(actor_id=self.id, peerid=peerid, config=self.config)
        return dbtrust.modify(baseuri=baseuri,
                              secret=secret,
                              desc=desc,
                              approved=approved,
                              verified=verified,
                              verification_token=verification_token,
                              peer_approved=peer_approved
                              )

    def create_reciprocal_trust(self, url, secret=None, desc='', relationship='', trust_type=''):
        """Creates a new reciprocal trust relationship locally and by requesting a relationship from a peer actor."""
        if len(url) == 0:
            return False
        if not secret or len(secret) == 0:
            return False
        res = self.get_peer_info(url)
        if not res or res["last_response_code"] < 200 or res["last_response_code"] >= 300:
            return False
        peer = res["data"]
        if not peer["id"] or not peer["type"] or len(peer["type"]) == 0:
            logging.info(
                "Received invalid peer info when trying to establish trust: " + url)
            return False
        if len(trust_type) > 0:
            if trust_type.lower() != peer["type"].lower():
                logging.info(
                    "Peer is of the wrong actingweb type: " + peer["type"])
                return False
        if not relationship or len(relationship) == 0:
            relationship = self.config.default_relationship
        # Create trust, so that peer can do a verify on the relationship (using
        # verification_token) when we request the relationship
        dbtrust = trust.Trust(actor_id=self.id, peerid=peer["id"], config=self.config)
        if not dbtrust.create(baseuri=url, secret=secret, peer_type=peer["type"],
                              relationship=relationship, approved=True,
                              verified=False, desc=desc):
            logging.warning("Trying to establish a new Reciprocal trust when peer relationship already exists (" +
                            peer["id"] + ")")
            return False
        # Since we are initiating the relationship, we implicitly approve it
        # It is not verified until the peer has verified us
        new_trust = dbtrust.get()
        params = {
            'baseuri': self.config.root + self.id,
            'id': self.id,
            'type': self.config.aw_type,
            'secret': secret,
            'desc': desc,
            'verify': new_trust["verification_token"],
        }
        requrl = url + '/trust/' + relationship
        data = json.dumps(params)
        logging.debug('Creating reciprocal trust at url(' +
                      requrl + ') and body (' + str(data) + ')')
        try:
            if self.config.env == 'appengine':
                self.config.module["urlfetch"].set_default_fetch_deadline(20)
                response = self.config.module["urlfetch"].fetch(url=requrl,
                                                                method=self.config.module["urlfetch"].POST,
                                                                payload=data,
                                                                headers={
                                                                    'Content-Type': 'application/json', }
                                                                )
            else:
                response = self.config.module["urlfetch"].post(
                    url=requrl,
                    data=data,
                    headers={
                        'Content-Type': 'application/json', }
                    )
            self.last_response_code = response.status_code
            self.last_response_message = response.content
        except (self.config.module["urlfetch"].UrlfetchException,
                self.config.module["urlfetch"].URLError,
                self.config.module["urlfetch"].Timeout,
                self.config.module["urlfetch"].TooManyRedirects):
            logging.debug(
                "Not able to create trust with peer, deleting my trust.")
            dbtrust.delete()
            return False
        if self.last_response_code == 201 or self.last_response_code == 202:
            # Reload the trust to check if approval was done
            mod_trust = trust.Trust(actor_id=self.id, peerid=peer["id"], config=self.config)
            mod_trust_data = mod_trust.get()
            if not mod_trust_data or len(mod_trust_data) == 0:
                logging.error(
                    "Couldn't find trust relationship after peer POST and verification")
                return False
            if self.last_response_code == 201:
                # Already approved by peer (probably auto-approved)
                # Do it direct on the trust (and not self.modifyTrustAndNotify) to avoid a callback
                # to the peer
                mod_trust.modify(peer_approved=True)
            return mod_trust.get()
        else:
            logging.debug(
                "Not able to create trust with peer, deleting my trust.")
            dbtrust.delete()
            return False

    def create_verified_trust(self, baseuri='', peerid=None, approved=False,
                              secret=None, verification_token=None, trust_type=None,
                              peer_approved=None, relationship=None, desc=''):
        """Creates a new trust when requested and call backs to initiating actor to verify relationship."""
        if not peerid or len(baseuri) == 0 or not relationship:
            return False
        requrl = baseuri + '/trust/' + relationship + '/' + self.id
        if not secret or len(secret) == 0:
            logging.debug('No secret received from requesting peer(' + peerid +
                          ') at url (' + requrl + '). Verification is not possible.')
            verified = False
        else:
            headers = {'Authorization': 'Bearer ' + secret,
                       }
            logging.debug('Verifying trust at requesting peer(' + peerid +
                          ') at url (' + requrl + ') and secret(' + secret + ')')
            try:
                if self.config.env == 'appengine':
                    self.config.module["urlfetch"].set_default_fetch_deadline(20)
                    response = self.config.module["urlfetch"].fetch(url=requrl,
                                                                    method=self.config.module["urlfetch"].GET,
                                                                    headers=headers)
                else:
                    response = self.config.module["urlfetch"].get(url=requrl, headers=headers)
                self.last_response_code = response.status_code
                self.last_response_message = response.content
                try:
                    logging.debug(
                        'Verifying trust response JSON:' + str(response.content))
                    data = json.loads(response.content.decode('utf-8', 'ignore'))
                    if data["verification_token"] == verification_token:
                        verified = True
                    else:
                        verified = False
                except ValueError:
                    logging.debug(
                        'No json body in response when verifying trust at url(' + requrl + ')')
                    verified = False
            except (self.config.module["urlfetch"].UrlfetchException,
                    self.config.module["urlfetch"].URLError,
                    self.config.module["urlfetch"].Timeout,
                    self.config.module["urlfetch"].TooManyRedirects):
                logging.debug(
                    'No response when verifying trust at url' + requrl + ')')
                verified = False
        new_trust = trust.Trust(actor_id=self.id, peerid=peerid, config=self.config)
        if not new_trust.create(baseuri=baseuri, secret=secret, peer_type=trust_type, approved=approved, 
                                peer_approved=peer_approved,
                                relationship=relationship, verified=verified, desc=desc):
            return False
        else:
            return new_trust.get()

    def delete_reciprocal_trust(self, peerid=None, delete_peer=False):
        """Deletes a trust relationship and requests deletion of peer's relationship as well."""
        failed_once = False  # For multiple relationships, this will be True if at least one deletion at peer failed
        success_once = False  # True if at least one relationship was deleted at peer
        if not peerid:
            rels = self.get_trust_relationships()
        else:
            rels = self.get_trust_relationships(peerid=peerid)
        for rel in rels:
            if delete_peer:
                url = rel["baseuri"] + '/trust/' + rel["relationship"] + '/' + self.id
                headers = {}
                if rel["secret"]:
                    headers = {'Authorization': 'Bearer ' + rel["secret"],
                               }
                logging.debug(
                    'Deleting reciprocal relationship at url(' + url + ')')
                try:
                    if self.config.env == 'appengine':
                        self.config.module["urlfetch"].set_default_fetch_deadline(20)
                        response = self.config.module["urlfetch"].fetch(url=url,
                                                                        method=self.config.module["urlfetch"].DELETE,
                                                                        headers=headers)
                    else:
                        response = self.config.module["urlfetch"].delete(url=url, headers=headers)
                except (self.config.module["urlfetch"].UrlfetchException,
                        self.config.module["urlfetch"].URLError,
                        self.config.module["urlfetch"].Timeout,
                        self.config.module["urlfetch"].TooManyRedirects):
                    logging.debug(
                        'Failed to delete reciprocal relationship at url(' + url + ')')
                    failed_once = True
                    continue
                if (response.status_code < 200 or response.status_code > 299) and response.status_code != 404:
                    logging.debug(
                        'Failed to delete reciprocal relationship at url(' + url + ')')
                    failed_once = True
                    continue
                else:
                    success_once = True
            if not self.subs_list:
                self.subs_list = subscription.Subscriptions(actor_id=self.id, config=self.config).fetch()
            # Delete this peer's subscriptions
            for sub in self.subs_list:
                if sub["peerid"] == rel["peerid"]:
                    logging.debug("Deleting subscription(" + sub["subscriptionid"] + ") as part of trust deletion.")
                    sub_obj = self.get_subscription_obj(peerid=sub["peerid"], subid=sub["subscriptionid"],
                                                        callback=sub["callback"])
                    sub_obj.delete()
            dbtrust = trust.Trust(actor_id=self.id, peerid=rel["peerid"], config=self.config)
            dbtrust.delete()
        if delete_peer and (not success_once or failed_once):
            return False
        return True

    def create_subscription(self, peerid=None, target=None, subtarget=None, resource=None, granularity=None,
                            subid=None, callback=False):
        new_sub = subscription.Subscription(
            actor_id=self.id, peerid=peerid, subid=subid, callback=callback, config=self.config)
        new_sub.create(target=target, subtarget=subtarget, resource=resource,
                       granularity=granularity)
        return new_sub.get()

    def create_remote_subscription(self, peerid=None, target=None, subtarget=None, resource=None, granularity=None):
        """Creates a new subscription at peerid."""
        if not peerid or not target:
            return False
        relationships = self.get_trust_relationships(peerid=peerid)
        if not relationships:
            return False
        peer = relationships[0]
        params = {
            'id': self.id,
            'target': target,
        }
        if subtarget:
            params['subtarget'] = subtarget
        if resource:
            params['resource'] = resource
        if granularity and len(granularity) > 0:
            params['granularity'] = granularity
        requrl = peer["baseuri"] + '/subscriptions/' + self.id
        data = json.dumps(params)
        headers = {'Authorization': 'Bearer ' + peer["secret"],
                   'Content-Type': 'application/json',
                   }
        try:
            logging.debug('Creating remote subscription at url(' +
                          requrl + ') with body (' + str(data) + ')')
            if self.config.env == 'appengine':
                self.config.module["urlfetch"].set_default_fetch_deadline(20)
                response = self.config.module["urlfetch"].fetch(url=requrl,
                                                                method=self.config.module["urlfetch"].POST,
                                                                payload=data,
                                                                headers=headers
                                                                )
            else:
                response = self.config.module["urlfetch"].post(url=requrl, data=data, headers=headers)
            self.last_response_code = response.status_code
            self.last_response_message = response.content
        except (self.config.module["urlfetch"].UrlfetchException,
                self.config.module["urlfetch"].URLError,
                self.config.module["urlfetch"].Timeout,
                self.config.module["urlfetch"].TooManyRedirects):
            return None
        try:
            logging.debug('Created remote subscription at url(' + requrl +
                          ') and got JSON response (' + str(response.content) + ')')
            data = json.loads(response.content.decode('utf-8', 'ignore'))
        except ValueError:
            return None
        if 'subscriptionid' in data:
            subid = data["subscriptionid"]
        else:
            return None
        if self.last_response_code == 201:
            self.create_subscription(peerid=peerid, target=target,
                                     subtarget=subtarget, resource=resource, granularity=granularity, subid=subid, 
                                     callback=True)
            if 'Location' in response.headers:
                return response.headers['Location']
            elif 'location' in response.headers:
                return response.headers['location']
        else:
            return None

    def get_subscriptions(self, peerid=None, target=None, subtarget=None, resource=None, callback=False):
        """Retrieves subscriptions from db."""
        if not self.id:
            return None
        if not self.subs_list:
            self.subs_list = subscription.Subscriptions(actor_id=self.id, config=self.config).fetch()
        ret = []
        for sub in self.subs_list:
            if not peerid or (peerid and sub["peerid"] == peerid):
                if not target or (target and sub["target"] == target):
                    if not subtarget or (subtarget and sub["subtarget"] == subtarget):
                        if not resource or (resource and sub["resource"] == resource):
                            if not callback or (callback and sub["callback"] == callback):
                                ret.append(sub)
        return ret

    def get_subscription(self, peerid=None, subid=None, callback=False):
        """Retrieves a single subscription identified by peerid and subid."""
        if not subid:
            return False
        return subscription.Subscription(
            actor_id=self.id, peerid=peerid, subid=subid, callback=callback, config=self.config).get()

    def get_subscription_obj(self, peerid=None, subid=None, callback=False):
        """Retrieves a single subscription identified by peerid and subid."""
        if not subid:
            return False
        return subscription.Subscription(
            actor_id=self.id, peerid=peerid, subid=subid, callback=callback, config=self.config)

    def delete_remote_subscription(self, peerid=None, subid=None):
        if not subid or not peerid:
            return False
        trust_rel = self.get_trust_relationship(peerid=peerid)
        if not trust_rel:
            return False
        sub = self.get_subscription(peerid=peerid, subid=subid)
        if not sub:
            sub = self.get_subscription(peerid=peerid, subid=subid, callback=True)
        if 'callback' not in sub or not sub["callback"]:
            url = trust_rel["baseuri"] + '/subscriptions/' + self.id + '/' + subid
        else:
            url = trust_rel["baseuri"] + '/callbacks/subscriptions/' + self.id + '/' + subid
        headers = {'Authorization': 'Bearer ' + trust_rel["secret"],
                   }
        try:
            logging.debug('Deleting remote subscription at url(' + url + ')')
            if self.config.env == 'appengine':
                self.config.module["urlfetch"].set_default_fetch_deadline(20)
                response = self.config.module["urlfetch"].fetch(url=url,
                                                                method=self.config.module["urlfetch"].DELETE,
                                                                headers=headers)
            else:
                response = self.config.module["urlfetch"].delete(url=url, headers=headers)
            self.last_response_code = response.status_code
            self.last_response_message = response.content
            if response.status_code == 204:
                return True
            else:
                logging.debug(
                    'Failed to delete remote subscription at url(' + url + ')')
                return False
        except (self.config.module["urlfetch"].UrlfetchException,
                self.config.module["urlfetch"].URLError,
                self.config.module["urlfetch"].Timeout,
                self.config.module["urlfetch"].TooManyRedirects):
            return False

    def delete_subscription(self, peerid=None, subid=None, callback=False):
        """Deletes a specified subscription"""
        if not subid:
            return False
        sub = subscription.Subscription(
            actor_id=self.id, peerid=peerid, subid=subid, callback=callback, config=self.config)
        return sub.delete()

    def callback_subscription(self, peerid=None, sub_obj=None, sub=None, diff=None, blob=None):
        if not peerid or not diff or not sub or not blob:
            logging.warning("Missing parameters in callbackSubscription")
            return
        if "granularity" in sub and sub["granularity"] == "none":
            return
        trust_rel = self.get_trust_relationship(peerid)
        if not trust_rel:
            return
        params = {
            'id': self.id,
            'subscriptionid': sub["subscriptionid"],
            'target': sub["target"],
            'sequence': diff["sequence"],
            'timestamp': str(diff["timestamp"]),
            'granularity': sub["granularity"],
        }
        if sub["subtarget"]:
            params['subtarget'] = sub["subtarget"]
        if sub["resource"]:
            params['resource'] = sub["resource"]
        if sub["granularity"] == "high":
            try:
                params['data'] = json.loads(blob)
            except (TypeError, ValueError, KeyError):
                params['data'] = blob
        if sub["granularity"] == "low":
            params['url'] = self.config.root + self.id + '/subscriptions/' + \
                trust_rel["peerid"] + '/' + sub["subscriptionid"] + '/' + str(diff["sequence"])
        requrl = trust_rel["baseuri"] + '/callbacks/subscriptions/' + self.id + '/' + sub["subscriptionid"]
        data = json.dumps(params)
        headers = {'Authorization': 'Bearer ' + trust_rel["secret"],
                   'Content-Type': 'application/json',
                   }
        try:
            logging.debug('Doing a callback on subscription at url(' +
                          requrl + ') with body(' + str(data) + ')')
            if self.config.env == 'appengine':
                self.config.module["urlfetch"].set_default_fetch_deadline(20)
                response = self.config.module["urlfetch"].fetch(url=requrl,
                                                                method=self.config.module["urlfetch"].POST,
                                                                payload=data.encode('utf-8'),
                                                                headers=headers
                                                                )
            else:
                response = self.config.module["urlfetch"].post(
                    url=requrl,
                    data=data.encode('utf-8'),
                    headers=headers
                    )
        except (self.config.module["urlfetch"].UrlfetchException,
                self.config.module["urlfetch"].URLError,
                self.config.module["urlfetch"].Timeout,
                self.config.module["urlfetch"].TooManyRedirects):
            logging.debug(
                'Peer did not respond to callback on url(' + requrl + ')')
            self.last_response_code = 0
            self.last_response_message = 'No response from peer for subscription callback'
            return
        self.last_response_code = response.status_code
        self.last_response_message = response.content
        if response.status_code == 204 and sub["granularity"] == "high":
            if not sub_obj:
                logging.warning("About to clear diff without having subobj set")
            else:
                sub_obj.clear_diff(diff["sequence"])

    def register_diffs(self, target=None, subtarget=None, resource=None, blob=None):
        """Registers a blob diff against all subscriptions with the correct target, subtarget, and resource.

            If resource is set, the blob is expected to be the FULL resource object, not a diff.
            """
        if blob is None or not target:
            return
        # Get all subscriptions, both with the specific subtarget/resource and those
        # without
        subs = self.get_subscriptions(
            target=target, subtarget=None, resource=None, callback=False)
        if not subs:
            subs = []
        if subtarget and resource:
            logging.debug("register_diffs() - blob(" + blob + "), target(" +
                          target + "), subtarget(" + subtarget + "), resource(" +
                          resource + "), # of subs(" + str(len(subs)) + ")")
        elif subtarget:
            logging.debug("register_diffs() - blob(" + blob + "), target(" +
                          target + "), subtarget(" + subtarget + 
                          "), # of subs(" + str(len(subs)) + ")")            
        else:
            logging.debug("register_diffs() - blob(" + blob + "), target(" +
                          target + "), # of subs(" + str(len(subs)) + ")")
        for sub in subs:
            # Skip the ones without correct subtarget
            if subtarget and sub["subtarget"] and sub["subtarget"] != subtarget:
                logging.debug("     - no match on subtarget, skipping...")
                continue
            # Skip the ones without correct resource
            if resource and sub["resource"] and sub["resource"] != resource:
                logging.debug("     - no match on resource, skipping...")
                continue
            sub_obj = self.get_subscription_obj(peerid=sub["peerid"], subid=sub["subscriptionid"])
            sub_obj_data = sub_obj.get()
            logging.debug("     - processing subscription(" + sub["subscriptionid"] +
                          ") for peer(" + sub["peerid"] + ") with target(" + 
                          sub_obj_data["target"] + ") subtarget(" + str(sub_obj_data["subtarget"] or '') +
                          ") and resource(" + str(sub_obj_data["resource"] or '') + ")")
            # Subscription with a resource, but this diff is on a higher level
            if (not resource or not subtarget) and sub_obj_data["subtarget"] and sub_obj_data["resource"]:
                # Create a json diff on the subpart that this subscription
                # covers
                try:
                    jsonblob = json.loads(blob)
                    if not subtarget:
                        subblob = json.dumps(jsonblob[sub_obj_data["subtarget"]][sub_obj_data["resource"]])
                    else:
                        subblob = json.dumps(jsonblob[sub_obj_data["resource"]])
                except (TypeError, ValueError, KeyError):
                    # The diff does not contain the resource
                    logging.debug("         - subscription has resource(" +
                                  sub_obj_data["resource"] + "), no matching blob found in diff")
                    continue
                logging.debug("         - subscription has resource(" +
                              sub_obj_data["resource"] + "), adding diff(" + subblob + ")")
                finblob = subblob
            # The diff is on the resource, but the subscription is on a 
            # higher level
            elif resource and not sub_obj_data["resource"]:
                # Since we have a resource, we know the blob is the entire resource, not a diff
                # If the subscription is for a sub-target, send [resource] = blob
                # If the subscription is for a target, send [subtarget][resource] = blob
                upblob = {}
                try:
                    jsonblob = json.loads(blob)
                    if not sub_obj_data["subtarget"]:
                        upblob[subtarget] = {}
                        upblob[subtarget][resource] = jsonblob
                    else:
                        upblob[resource] = jsonblob
                except (TypeError, ValueError, KeyError):
                    if not sub_obj_data["subtarget"]:
                        upblob[subtarget] = {}
                        upblob[subtarget][resource] = blob
                    else:
                        upblob[resource] = blob
                finblob = json.dumps(upblob)
                logging.debug("         - diff has resource(" + resource +
                              "), subscription has not, adding diff(" + finblob + ")")
            # Subscriptions with subtarget, but this diff is on a higher level
            elif not subtarget and sub_obj_data["subtarget"]:
                # Create a json diff on the subpart that this subscription
                # covers
                subblob = None
                try:
                    jsonblob = json.loads(blob)
                    subblob = json.dumps(jsonblob[sub_obj_data["subtarget"]])
                except (TypeError, ValueError, KeyError):
                    # The diff blob does not contain the subtarget
                    pass
                logging.debug("         - subscription has subtarget(" +
                              sub_obj_data["subtarget"] + "), adding diff(" + subblob + ")")
                finblob = subblob
            # The diff is on the subtarget, but the subscription is on the
            # higher level
            elif subtarget and not sub_obj_data["subtarget"]:
                # Create a data["subtarget"] = blob diff to give correct level
                # of diff to subscriber
                upblob = {}
                try:
                    jsonblob = json.loads(blob)
                    upblob[subtarget] = jsonblob
                except (TypeError, ValueError, KeyError):
                    upblob[subtarget] = blob
                finblob = json.dumps(upblob)
                logging.debug("         - diff has subtarget(" + subtarget +
                              "), subscription has not, adding diff(" + finblob + ")")
            else:
                # The diff is correct for the subscription
                logging.debug(
                              "         - exact target/subtarget match, adding diff(" + blob + ")")
                finblob = blob
            diff = sub_obj.add_diff(blob=finblob)
            if not diff:
                logging.warning("Failed when registering a diff to subscription (" +
                                sub["subscriptionid"] + "). Will not send callback.")
            else:
                if self.config.module["deferred"]:
                    self.config.module["deferred"].defer(self.callback_subscription, peerid=sub["peerid"],
                                                         sub_obj=sub_obj,
                                                         sub=sub_obj_data, diff=diff, blob=finblob)
                else:
                    self.callback_subscription(peerid=sub["peerid"], sub_obj=sub_obj,
                                               sub=sub_obj_data, diff=diff, blob=finblob)


class Actors(object):
    """ Handles all actors
    """

    def fetch(self):
        if not self.list:
            return False
        if self.actors is not None:
            return self.actors
        self.actors = self.list.fetch()
        return self.actors

    def __init__(self, config=None):
        self.config = config
        self.list = self.config.DbActor.DbActorList()
        self.actors = None
        self.fetch()



import json

from actingweb import auth
from actingweb.handlers import base_handler


class RootHandler(base_handler.BaseHandler):

    def get(self, actor_id):
        if self.request.get('_method') == 'DELETE':
            self.delete(actor_id)
            return
        (myself, check) = auth.init_actingweb(appreq=self, actor_id=actor_id, path='', subpath='',
                                              config=self.config)
        if not myself or check.response["code"] != 200:
            return
        if not check.check_authorisation(path='/', method='GET'):
            self.response.set_status(403)
            return
        pair = {
            'id': myself.id,
            'creator': myself.creator,
            'passphrase': myself.passphrase,
        }
        trustee_root = myself.store.trustee_root
        if self.config.migrate_2_5_0 and not trustee_root:
            trustee_root = myself.property.trustee_root
            if trustee_root:
                myself.property.trustee_root = None
                myself.store.trustee_root = trustee_root
        if trustee_root and len(trustee_root) > 0:
            pair['trustee_root'] = trustee_root
        out = json.dumps(pair)
        self.response.write(out.encode('utf-8'))
        self.response.headers["Content-Type"] = "application/json"
        self.response.set_status(200)

    def delete(self, actor_id):
        (myself, check) = auth.init_actingweb(appreq=self,
                                              actor_id=actor_id, path='', subpath='', config=self.config)
        if not myself or check.response["code"] != 200:
            return
        if not check.check_authorisation(path='/', method='DELETE'):
            self.response.set_status(403)
            return
        self.on_aw.delete_actor()
        myself.delete()
        self.response.set_status(204)
        return



from __future__ import absolute_import
from builtins import str
from builtins import object
import json
from actingweb import trust
import logging
try:
    from urllib.parse import urlencode as urllib_urlencode
except ImportError:
    from urllib import urlencode as urllib_urlencode


class AwProxy(object):
    """ Proxy to other trust peers to execute RPC style calls

    Initialise with either trust_target to target a specific
    existing trust or use peer_target for simplicity to use
    the trust established with the peer. 
    """

    def __init__(self, trust_target=None, peer_target=None, config=None):
        self.config = config
        self.last_response_code = 0
        self.last_response_message = 0
        self.last_location = None
        if trust_target and trust_target.trust:
            self.trust = trust_target
            self.actorid = trust_target.id
        elif peer_target and peer_target["id"]:
            self.actorid = peer_target["id"]
            self.trust = None
            if peer_target["peerid"]:
                self.trust = trust.Trust(
                    actor_id=self.actorid,
                    peerid=peer_target["peerid"],
                    config=self.config).get()
                if not self.trust or len(self.trust) == 0:
                    self.trust = None

    def get_resource(self, path=None, params=None):
        if not path or len(path) == 0:
            return None
        if not params:
            params = {}
        if not self.trust or not self.trust["baseuri"] or not self.trust["secret"]:
            return None
        url = self.trust["baseuri"].strip('/') + '/' + path.strip('/')
        if params:
            url = url + '?' + urllib_urlencode(params)
        headers = {'Authorization': 'Bearer ' + self.trust["secret"],
                   }
        logging.debug(
            'Getting trust peer resource at (' + url + ')')
        try:
            if self.config.env == 'appengine':
                response = self.config.module["urlfetch"].fetch(
                    url=url,
                    method=self.config.module["urlfetch"].GET,
                    headers=headers
                    )
            else:
                response = self.config.module["urlfetch"].get(
                    url=url,
                    headers=headers
                    )
            self.last_response_code = response.status_code
            self.last_response_message = response.content
        except (self.config.module["urlfetch"].UrlfetchException,
                self.config.module["urlfetch"].URLError,
                self.config.module["urlfetch"].Timeout,
                self.config.module["urlfetch"].TooManyRedirects):
            logging.debug('Not able to get peer resource')
            self.last_response_code = 408
            return {
                'error': {
                    'code': 408,
                    'message': 'Unable to communciate with trust peer service.',
                },
            }
        logging.debug('Get trust peer resource POST response:(' +
                      str(response.status_code) + ') ' + str(response.content))
        if response.status_code < 200 or response.status_code > 299:
            logging.info('Not able to get trust peer resource.')
        try:
            result = json.loads(response.content.decode('utf-8', 'ignore'))
        except (TypeError, ValueError, KeyError):
            logging.debug("Not able to parse response when getting resource at(" + url + ")")
            result = {}
        return result

    def create_resource(self, path=None, params=None):
        if not path or len(path) == 0:
            return None
        if not params:
            params = {}
        if not self.trust or not self.trust["baseuri"] or not self.trust["secret"]:
            return None
        data = json.dumps(params)
        headers = {'Authorization': 'Bearer ' + self.trust["secret"],
                   'Content-Type': 'application/json',
                   }
        url = self.trust["baseuri"].strip('/') + '/' + path.strip('/')
        logging.debug(
            'Creating trust peer resource at (' + url + ') with data(' +
            str(data) + ')')
        try:
            if self.config.env == 'appengine':
                response = self.config.module["urlfetch"].fetch(
                    url=url,
                    method=self.config.module["urlfetch"].POST,
                    payload=data,
                    headers=headers
                    )
            else:
                response = self.config.module["urlfetch"].fetch(
                    url=url,
                    data=data,
                    headers=headers
                    )
            self.last_response_code = response.status_code
            self.last_response_message = response.content
        except (self.config.module["urlfetch"].UrlfetchException,
                self.config.module["urlfetch"].URLError,
                self.config.module["urlfetch"].Timeout,
                self.config.module["urlfetch"].TooManyRedirects):
            logging.debug('Not able to create new peer resource')
            self.last_response_code = 408
            return {
                'error': {
                    'code': 408,
                    'message': 'Unable to communciate with trust peer service.',
                },
            }
        if 'Location' in response.headers:
            self.last_location = response.headers['Location']
        else:
            self.last_location = None
        logging.debug('Create trust peer resource POST response:(' +
                      str(response.status_code) + ') ' + str(response.content))
        if response.status_code < 200 or response.status_code > 299:
            logging.warning('Not able to create new trust peer resource.')
        try:
            result = json.loads(response.content.decode('utf-8', 'ignore'))
        except (TypeError, ValueError, KeyError):
            logging.debug("Not able to parse response when creating resource at(" + url + ")")
            result = {}
        return result

    def change_resource(self, path=None, params=None):
        if not path or len(path) == 0:
            return None
        if not params:
            params = {}
        if not self.trust or not self.trust["baseuri"] or not self.trust["secret"]:
            return None
        data = json.dumps(params)
        headers = {'Authorization': 'Bearer ' + self.trust["secret"],
                   'Content-Type': 'application/json',
                   }
        url = self.trust["baseuri"].strip('/') + '/' + path.strip('/')
        logging.debug(
            'Changing trust peer resource at (' + url + ') with data(' +
            str(data) + ')')
        try:
            if self.config.env == 'appengine':
                response = self.config.module["urlfetch"].fetch(
                    url=url,
                    method=self.config.module["urlfetch"].PUT,
                    payload=data,
                    headers=headers
                    )
            else:
                response = self.config.module["urlfetch"].put(
                    url=url,
                    data=data,
                    headers=headers
                )
            self.last_response_code = response.status_code
            self.last_response_message = response.content
        except (self.config.module["urlfetch"].UrlfetchException,
                self.config.module["urlfetch"].URLError,
                self.config.module["urlfetch"].Timeout,
                self.config.module["urlfetch"].TooManyRedirects):
            logging.debug('Not able to change peer resource')
            self.last_response_code = 408
            return {
                'error': {
                    'code': 408,
                    'message': 'Unable to communciate with trust peer service.',
                },
            }
        logging.debug('Change trust peer resource PUT response:(' +
                      str(response.status_code) + ') ' + str(response.content))
        if response.status_code < 200 or response.status_code > 299:
            logging.warning('Not able to change trust peer resource.')
        try:
            result = json.loads(response.content.decode('utf-8', 'ignore'))
        except (TypeError, ValueError, KeyError):
            logging.debug("Not able to parse response when changing resource at(" + url + ")")
            result = {}
        return result

    def delete_resource(self, path=None):
        if not path or len(path) == 0:
            return None
        if not self.trust or not self.trust["baseuri"] or not self.trust["secret"]:
            return None
        headers = {'Authorization': 'Bearer ' + self.trust["secret"],
                   }
        url = self.trust["baseuri"].strip('/') + '/' + path.strip('/')
        logging.debug(
            'Deleting trust peer resource at (' + url + ')')
        try:
            if self.config.env == 'appengine':
                response = self.config.module["urlfetch"].fetch(
                    url=url,
                    method=self.config.module["urlfetch"].DELETE,
                    headers=headers
                    )
            else:
                response = self.config.module["urlfetch"].delete(
                    url=url,
                    headers=headers
                    )
            self.last_response_code = response.status_code
            self.last_response_message = response.content
        except (self.config.module["urlfetch"].UrlfetchException,
                self.config.module["urlfetch"].URLError,
                self.config.module["urlfetch"].Timeout,
                self.config.module["urlfetch"].TooManyRedirects):
            logging.debug('Not able to delete peer resource')
            self.last_response_code = 408
            return {
                'error': {
                    'code': 408,
                    'message': 'Unable to communciate with trust peer service.',
                },
            }
        logging.debug('Delete trust peer resource POST response:(' +
                      str(response.status_code) + ') ' + str(response.content))
        if response.status_code < 200 or response.status_code > 299:
            logging.warning('Not able to delete trust peer resource.')
            return False
        return True



import json

from actingweb import auth
from actingweb.handlers import base_handler


class ResourcesHandler(base_handler.BaseHandler):

    def get(self, actor_id, name):
        (myself, check) = auth.init_actingweb(appreq=self,
                                              actor_id=actor_id,
                                              path='resources',
                                              subpath=name,
                                              config=self.config)
        if not myself or check.response["code"] != 200:
            return
        if not check.check_authorisation(path='resources', subpath=name, method='GET'):
            self.response.set_status(403)
            return
        pair = self.on_aw.get_resources(name=name)
        if pair and any(pair): 
            out = json.dumps(pair)
            self.response.write(out.encode('utf-8'))
            self.response.headers["Content-Type"] = "application/json"
            self.response.set_status(200)
        else:
            self.response.set_status(404)

    def delete(self, actor_id, name):
        (myself, check) = auth.init_actingweb(appreq=self,
                                              actor_id=actor_id,
                                              path='resources',
                                              subpath=name,
                                              config=self.config)
        if not myself or check.response["code"] != 200:
            return
        if not check.check_authorisation(path='resources', subpath=name, method='DELETE'):
            self.response.set_status(403)
            return
        pair = self.on_aw.delete_resources(name=name)
        if pair:
            if 100 <= pair <= 999:
                return
            if any(pair): 
                out = json.dumps(pair)
                self.response.write(out.encode('utf-8'))
                self.response.headers["Content-Type"] = "application/json"
                self.response.set_status(200)
        else:
            self.response.set_status(404)

    def put(self, actor_id, name):
        (myself, check) = auth.init_actingweb(appreq=self,
                                              actor_id=actor_id,
                                              path='resources',
                                              subpath=name,
                                              config=self.config)
        if not myself or check.response["code"] != 200:
            return
        if not check.check_authorisation(path='resources', subpath=name, method='PUT'):
            self.response.set_status(403)
            return
        try:
            params = json.loads(self.request.body.decode('utf-8', 'ignore'))
        except (TypeError, ValueError, KeyError):
            self.response.set_status(400, "Error in json body")
            return
        pair = self.on_aw.put_resources(name=name, params=params)
        if pair:
            if 100 <= pair <= 999:
                return
            if any(pair):
                out = json.dumps(pair)
                self.response.write(out.encode('utf-8'))
                self.response.headers["Content-Type"] = "application/json"
                self.response.set_status(200)
        else:
            self.response.set_status(404)

    def post(self, actor_id, name):
        (myself, check) = auth.init_actingweb(appreq=self,
                                              actor_id=actor_id,
                                              path='resources',
                                              subpath=name,
                                              config=self.config)
        if not myself or check.response["code"] != 200:
            return
        if not check.check_authorisation(path='resources', subpath=name, method='POST'):
            self.response.set_status(403)
            return
        try:
            params = json.loads(self.request.body.decode('utf-8', 'ignore'))
        except (TypeError, ValueError, KeyError):
            self.response.set_status(400, "Error in json body")
            return
        pair = self.on_aw.post_resources(name=name, params=params)
        if pair:
            if 100 <= pair <= 999:
                return
            if any(pair):
                out = json.dumps(pair)
                self.response.write(out.encode('utf-8'))
                self.response.headers["Content-Type"] = "application/json"
                self.response.set_status(201, 'Created')
        else:
            self.response.set_status(404)



from builtins import object
import uuid
import binascii
import logging
import importlib
import os


class Config(object):

    def __init__(self, **kwargs):
        #########
        # Basic settings for this app
        #########
        # Values that can be changed as part of instantiating config
        self.fqdn = "actingwebdemo-dev.appspot.com"  # The host and domain, i.e. FQDN, of the URL
        self.proto = "https://"  # http or https
        self.env = ''
        self.database = 'dynamodb'
        # Turn on the /www path
        self.ui = True
        # Enable /devtest path for test purposes, MUST be False in production
        self.devtest = True
        # Enable migrate if you want to turn on code that enables migration from one version to another
        # 2.4.4 : migrate oauth properties and email to internal attributes (myself.store)
        self.migrate_2_5_0 = True
        # Will enforce unique creator field across all actors
        self.unique_creator = False
        # Use "email" internal value to set creator value (after creation and property set)
        self.force_email_prop_as_creator = True
        # basic or oauth: basic for creator + bearer tokens
        self.www_auth = "basic"
        self.logLevel = logging.DEBUG
        # Change to WARN for production, DEBUG for debugging, and INFO for normal testing
        #########
        # Configurable ActingWeb settings for this app
        #########
        self.aw_type = "urn:actingweb:actingweb.org:gae-demo"  # The app type this actor implements
        self.desc = "GAE Demo actor: "                      # A human-readable description for this specific actor
        self.specification = ""                             # URL to a RAML/Swagger etc definition if available
        self.version = "1.0"                                # A version number for this app
        self.info = "http://actingweb.org/"                 # Where can more info be found
        #########
        # Trust settings for this app
        #########
        self.default_relationship = "associate"  # Default relationship if not specified
        self.auto_accept_default_relationship = False  # True if auto-approval
        #########
        # Known and trusted ActingWeb actors
        #########
        self.actors = {
            '<SHORTTYPE>': {
                'type': 'urn:<ACTINGWEB_TYPE>',
                'factory': '<ROOT_URI>',
                'relationship': 'friend',                   # associate, friend, partner, admin
                },
        }
        #########
        # OAuth settings for this app, fill in if OAuth is used
        #########
        self.oauth = {
            'client_id': "",                                # An empty client_id turns off oauth capabilities
            'client_secret': "",
            'redirect_uri': self.proto + self.fqdn + "/oauth",
            'scope': "",
            'auth_uri': "",
            'token_uri': "",
            'response_type': "code",
            'grant_type': "authorization_code",
            'refresh_type': "refresh_token",
        }
        self.bot = {
            'token': '',
            'email': '',
        }
        # List of paths and their access levels
        # Matching is done top to bottom stopping at first match (role, path)
        # If no match is found on path with the correct role, access is rejected
        # <type> and <id> are used as templates for trust types and ids
        self.access = [
            # (role, path, method, access), e.g. ('friend', '/properties', '', 'a')
            # Roles: creator, trustee, associate, friend, partner, admin, any (i.e. authenticated),
            #        owner (i.e. trust peer owning the entity)
            #        + any other new role for this app
            # Methods: GET, POST, PUT, DELETE
            # Access: a (allow) or r (reject)
            ('', 'meta', 'GET', 'a'),                       # Allow GET to anybody without auth
            ('', 'oauth', '', 'a'),                         # Allow any method to anybody without auth
            ('owner', 'callbacks/subscriptions', 'POST', 'a'),   # Allow owners on subscriptions
            ('', 'callbacks', '', 'a'),                     # Allow anybody callbacks witout auth
            ('creator', 'www', '', 'a'),                    # Allow only creator access to /www
            ('creator', 'properties', '', 'a'),             # Allow creator access to /properties
            ('associate', 'properties', 'GET', 'a'),        # Allow GET only to associate
            ('friend', 'properties', '', 'a'),              # Allow friend/partner/admin all
            ('partner', 'properties', '', 'a'),
            ('admin', 'properties', '', 'a'),
            ('creator', 'resources', '', 'a'),
            ('friend', 'resources', '', 'a'),               # Allow friend/partner/admin all
            ('partner', 'resources', '', 'a'),
            ('admin', 'resources', '', 'a'),
            ('', 'trust/<type>', 'POST', 'a'),              # Allow unauthenticated POST
            ('owner', 'trust/<type>/<id>', '', 'a'),        # Allow trust peer full access
            ('creator', 'trust', '', 'a'),                  # Allow access to all to
            ('trustee', 'trust', '', 'a'),                  # creator/trustee/admin
            ('admin', 'trust', '', 'a'),
            ('owner', 'subscriptions', '', 'a'),             # Owner can create++ own subscriptions
            ('friend', 'subscriptions/<id>', '', 'a'),       # Owner can create subscriptions
            ('creator', 'subscriptions', '', 'a'),           # Creator can do everything
            ('trustee', 'subscriptions', '', 'a'),           # Trustee can do everything
            ('creator', '/', '', 'a'),                       # Root access for actor
            ('trustee', '/', '', 'a'),
            ('admin', '/', '', 'a'),
        ]
        # Pick up the config variables
        for k, v in kwargs.items():
            self.__setattr__(k, v)
        if self.database == 'gae':
            self.env = 'appengine'
        elif self.database == 'dynamodb':
            self.env = 'aws'
        if self.logLevel == "DEBUG":
            self.logLevel = logging.DEBUG
        elif self.logLevel == "WARN":
            self.logLevel = logging.WARN
        elif self.logLevel == "INFO":
            self.logLevel = logging.INFO
        else:
            self.logLevel = logging.DEBUG
        if 'myself' not in self.actors:
            # Add myself as a known type
            self.actors['myself'] = {
                'type': self.aw_type,
                'factory': self.proto + self.fqdn + '/',
                'relationship': 'friend',  # associate, friend, partner, admin
            }
        # Dynamically load all the database modules
        self.DbActor = importlib.import_module("actingweb" + ".db_" + self.database + ".db_actor")
        self.DbPeerTrustee = importlib.import_module("actingweb" + ".db_" + self.database + ".db_peertrustee")
        self.DbProperty = importlib.import_module("actingweb" + ".db_" + self.database + ".db_property")
        self.DbAttribute = importlib.import_module("actingweb" + ".db_" + self.database + ".db_attribute")
        self.DbSubscription = importlib.import_module("actingweb" + ".db_" + self.database + ".db_subscription")
        self.DbSubscriptionDiff = importlib.import_module("actingweb" + ".db_" + self.database +
                                                          ".db_subscription_diff")
        self.DbTrust = importlib.import_module("actingweb" + ".db_" + self.database + ".db_trust")
        self.module = {}
        if self.env == 'appengine':
            self.module["deferred"] = importlib.import_module(".deferred", "google.appengine.api")
            self.module["urlfetch"] = importlib.import_module(".urlfetch", "google.appengine.ext")
        else:
            self.module["deferred"] = None
            self.module["urlfetch"] = importlib.import_module("urlfetch")
        #########
        # ActingWeb settings for this app
        #########
        self.aw_version = "1.0"                             # This app follows the actingweb specification specified
        self.aw_supported = "www,oauth,callbacks,trust,onewaytrust,subscriptions," \
                            "actions,resources,methods,sessions,nestedproperties" # This app supports these options
        self.aw_formats = "json"                            # These are the supported formats
        #########
        # Only touch the below if you know what you are doing
        #########
        if self.env == 'appengine':
            logging.getLogger().handlers[0].setLevel(self.logLevel)  # Hack to get access to GAE logger
        else:
            logging.basicConfig(level=self.logLevel)
            # Turn off debugging for pynamodb and botocore, too noisy
            if self.logLevel == logging.DEBUG:
                log = logging.getLogger("pynamodb")
                log.setLevel(logging.INFO)
                log.propagate = True
                log = logging.getLogger("botocore")
                log.setLevel(logging.INFO)
                log.propagate = True
        self.root = self.proto + self.fqdn + "/"            # root URI used to identity actor externally
        self.auth_realm = self.fqdn                         # Authentication realm used in Basic auth

    @staticmethod
    def new_uuid(seed):
        return uuid.uuid5(uuid.NAMESPACE_URL, str(seed)).hex

    @staticmethod
    def new_token(length=40):
        tok = binascii.hexlify(os.urandom(int(length // 2)))
        return tok.decode('utf-8')



from builtins import object


class InternalStore(object):
    """ Access to internal attributes using .prop notation
    """

    def __init__(self, actor_id=None, config=None, bucket=None):
        if not bucket:
            bucket = '_internal'
        self._db = Attributes(actor_id=actor_id, bucket=bucket, config=config)
        d = self._db.get_bucket()
        if d:
            for k, v in d.items():
                self.__setattr__(k, v.get('data'))
        self.__initialised = True

    def __getitem__(self, k):
        return self.__getattr__(k)

    def __setitem__(self, k, v):
        return self.__setattr__(k, v)

    def __setattr__(self, k, v):
        if '_InternalStore__initialised' not in self.__dict__:
            return object.__setattr__(self, k, v)
        if k is None:
            raise ValueError
        if v is None:
            self.__dict__['_db'].delete_attr(name=k)
            if k in self.__dict__:
                self.__delattr__(k)
        else:
            self.__dict__[k] = v
            self.__dict__['_db'].set_attr(name=k, data=v)

    def __getattr__(self, k):
        try:
            return self.__dict__[k]
        except KeyError:
            return None


class Attributes(object):
    """
        Attributes is the main entity keeping an attribute.

        It needs to be initalized at object creation time.

    """

    def get_bucket(self):
        """ Retrieves the attribute bucket from the database """
        if not self.data or len(self.data) == 0:
            self.data = self.dbprop.get_bucket(actor_id=self.actor_id, bucket=self.bucket)
        return self.data

    def get_attr(self, name=None):
        """ Retrieves a single attribute """
        if not name:
            return None
        if name not in self.data:
            self.data[name] = self.dbprop.get_attr(actor_id=self.actor_id, bucket=self.bucket, name=name)
        return self.data[name]

    def set_attr(self, name=None, data=None, timestamp=None):
        """ Sets new data for this attribute """
        if not self.actor_id or not self.bucket:
            return False
        if name not in self.data or self.data[name] is None:
            self.data[name] = {}
        self.data[name]["data"] = data
        self.data[name]["timestamp"] = timestamp
        return self.dbprop.set_attr(
            actor_id=self.actor_id,
            bucket=self.bucket,
            name=name,
            data=data,
            timestamp=timestamp
        )

    def delete_attr(self, name=None):
        if not name:
            return False
        if 'name' in self.data:
            del self.data[name]
        return self.dbprop.delete_attr(actor_id=self.actor_id, bucket=self.bucket, name=name)

    def delete_bucket(self):
        """ Deletes the attribute bucket in the database """
        if not self.dbprop:
            return False
        if self.dbprop.delete_bucket(actor_id=self.actor_id, bucket=self.bucket):
            self.dbprop = self.config.DbAttribute.DbAttribute()
            self.data = {}
            return True
        else:
            return False

    def __init__(self,  actor_id=None, bucket=None, config=None):
        """ A attribute must be initialised with actor_id and bucket
        """
        self.config = config
        self.dbprop = self.config.DbAttribute.DbAttribute()
        self.bucket = bucket
        self.actor_id = actor_id
        self.data = {}
        if actor_id and bucket and len(bucket) > 0 and config:
            self.get_bucket()


class Buckets(object):
    """ Handles all attribute buckets of a specific actor_id

        Access the attributes
        in .props as a dictionary
    """

    def fetch(self):
        if not self.actor_id:
            return False
        return self.list.fetch(actor_id=self.actor_id)

    def fetch_timestamps(self):
        if not self.actor_id:
            return False
        return self.list.fetch_timestamps(actor_id=self.actor_id)

    def delete(self):
        if not self.list:
            return False
        self.list.delete(actor_id=self.actor_id)
        self.list = self.config.DbAttribute.DbAttributeBucketList()
        return True

    def __init__(self,  actor_id=None, config=None):
        """ attributes must always be initialised with an actor_id """
        self.config = config
        if not actor_id:
            self.list = None
            return
        self.list = self.config.DbAttribute.DbAttributeBucketList()
        self.actor_id = actor_id



from builtins import object


class PropertyStore(object):

    def __init__(self, actor_id=None, config=None):
        self._actor_id = actor_id
        self._config = config
        self.__initialised = True

    def __getitem__(self, k):
        return self.__getattr__(k)

    def __setitem__(self, k, v):
        return self.__setattr__(k, v)

    def __setattr__(self, k, v):
        if '_PropertyStore__initialised' not in self.__dict__:
            return object.__setattr__(self, k, v)
        if v is None:
            if k in self.__dict__:
                self.__delattr__(k)
        else:
            self.__dict__[k] = v
        # Re-init property to avoid overwrite
        self.__dict__['_db'] = self.__dict__['_config'].DbProperty.DbProperty()
        # set() will retrieve an attribute and delete it if value = None
        self.__dict__['_db'].set(actor_id=self.__dict__['_actor_id'], name=k, value=v)

    def __getattr__(self, k):
        try:
            return self.__dict__[k]
        except KeyError:
            self.__dict__['_db'] = self.__dict__['_config'].DbProperty.DbProperty()
            self.__dict__[k] = self.__dict__['_db'].get(actor_id=self.__dict__['_actor_id'], name=k)
            return self.__dict__[k]


class Property(object):
    """
        property is the main entity keeping a property.

        It needs to be initalised at object creation time.

    """

    def get(self):
        """ Retrieves the property from the database """
        if not self.dbprop:
            # New property after a delete()
            self.dbprop = self.config.DbProperty.DbProperty()
            self.value = None
        self.value = self.dbprop.get(actor_id=self.actor_id, name=self.name)
        return self.value

    def set(self, value):
        """ Sets a new value for this property """
        if not self.dbprop:
            # New property after a delete()
            self.dbprop = self.config.DbProperty.DbProperty()
        if not self.actor_id or not self.name:
            return False
        # Make sure we have made a dip in db to avoid two properties
        # with same name
        db_value = self.dbprop.get(actor_id=self.actor_id, name=self.name)
        if db_value == value:
            return True
        self.value = value
        return self.dbprop.set(actor_id=self.actor_id, name=self.name, value=value)

    def delete(self):
        """ Deletes the property in the database """
        if not self.dbprop:
            return
        if self.dbprop.delete():
            self.value = None
            self.dbprop = None
            return True
        else:
            return False

    def get_actor_id(self):
        return self.actor_id

    def __init__(self, actor_id=None, name=None, value=None, config=None):
        """ A property must be initialised with actor_id and name or
            name and value (to find an actor's property of a certain value)
        """
        self.config = config
        self.dbprop = self.config.DbProperty.DbProperty()
        self.name = name
        if not actor_id and name and len(name) > 0 and value and len(value) > 0:
            self.actor_id = self.dbprop.get_actor_id_from_property(name=name,
                                                                   value=value)
            if not self.actor_id:
                return
            self.value = value
        else:
            self.actor_id = actor_id
            self.value = None
            if name and len(name) > 0:
                self.get()


class Properties(object):
    """ Handles all properties of a specific actor_id

        Access the properties
        in .props as a dictionary
    """

    def fetch(self):
        if not self.actor_id:
            return False
        if not self.list:
            return False
        if self.props is not None:
            return self.props
        self.props = self.list.fetch(actor_id=self.actor_id)
        return self.props

    def delete(self):
        if not self.list:
            self.fetch()
        if not self.list:
            return False
        self.list.delete()
        return True

    def __init__(self,  actor_id=None, config=None):
        """ Properties must always be initialised with an actor_id """
        self.config = config
        if not actor_id:
            self.list = None
            return
        self.list = self.config.DbProperty.DbPropertyList()
        self.actor_id = actor_id
        self.props = None
        self.fetch()



import json
import logging
import copy
from actingweb import auth
from actingweb.handlers import base_handler


def merge_dict(d1, d2):
    """ Modifies d1 in-place to contain values from d2.

    If any value in d1 is a dictionary (or dict-like), *and* the corresponding
    value in d2 is also a dictionary, then merge them in-place.
    Thanks to Edward Loper on stackoverflow.com
    """
    for k, v2 in list(d2.items()):
        v1 = d1.get(k)  # returns None if v1 has no value for this key
        if isinstance(v1, dict) and isinstance(v2, dict):
            merge_dict(v1, v2)
        else:
            d1[k] = v2


def delete_dict(d1, path):
    """ Deletes path (an array of strings) in d1 dict.

    d1 is modified to no longer contain the attr/value pair
    or dict that is specified by path.
    """
    if not d1:
        # logging.debug('Path not found')
        return False
    # logging.debug('d1: ' + json.dumps(d1))
    # logging.debug('path: ' + str(path))
    if len(path) > 1 and path[1] and len(path[1]) > 0:
        return delete_dict(d1.get(path[0]), path[1:])
    if len(path) == 1 and path[0] and path[0] in d1:
        # logging.debug('Deleting d1[' + path[0] + ']')
        try:
            del d1[path[0]]
            return True
        except KeyError:
            return False
    return False


class PropertiesHandler(base_handler.BaseHandler):

    def get(self, actor_id, name):
        if self.request.get('_method') == 'PUT':
            self.put(actor_id, name)
            return
        if self.request.get('_method') == 'DELETE':
            self.delete(actor_id, name)
            return
        (myself, check) = auth.init_actingweb(appreq=self,
                                              actor_id=actor_id, path='properties', subpath=name,
                                              config=self.config)
        if not myself or check.response["code"] != 200:
            return
        if not name:
            path = {0: None}
        else:
            path = name.split('/')
            name = path[0]
        if not check.check_authorisation(path='properties', subpath=name, method='GET'):
            self.response.set_status(403)
            return
        # if name is not set, this request URI was the properties root
        if not name:
            self.listall(myself)
            return
        lookup = myself.property[name]
        if not lookup:
            self.response.set_status(404, "Property not found")
            return
        try:
            jsonblob = json.loads(lookup)
            try:
                out = jsonblob
                if len(path) > 1:
                    del path[0]
                    for p in path:
                        out = out[p]
                out = self.on_aw.get_properties(path=path, data=out)
                if out is None:
                    self.response.set_status(404)
                    return
                out = json.dumps(out)
            except (TypeError, ValueError, KeyError):
                self.response.set_status(404)
                return
            out = out.encode('utf-8')
        except (TypeError, ValueError, KeyError):
                out = lookup
        self.response.set_status(200, "Ok")
        self.response.headers["Content-Type"] = "application/json"
        self.response.write(out)

    def listall(self, myself):
        properties = myself.get_properties()
        if not properties or len(properties) == 0:
            self.response.set_status(404, "No properties")
            return
        pair = dict()
        for name, value in list(properties.items()):
            try:
                js = json.loads(value)
                pair[name] = js
            except ValueError:
                pair[name] = value
        pair = self.on_aw.get_properties(path=None, data=pair)
        if pair is None:
            self.response.set_status(404)
            return
        out = json.dumps(pair)
        self.response.write(out.encode('utf-8'))
        self.response.headers["Content-Type"] = "application/json"
        return

    def put(self, actor_id, name):
        (myself, check) = auth.init_actingweb(appreq=self,
                                              actor_id=actor_id, path='properties', subpath=name,
                                              config=self.config)
        if not myself or check.response["code"] != 200:
            return
        resource = None
        if not name:
            path = []
        else:
            path = name.split('/')
            name = path[0]
            if len(path) >= 2 and len(path[1]) > 0:
                resource = path[1]
        if not check.check_authorisation(path='properties', subpath=name, method='PUT'):
            self.response.set_status(403)
            return
        body = self.request.body.decode('utf-8', 'ignore')
        if len(path) == 1:
            old = myself.property[name]
            try:
                old = json.loads(old)
            except (TypeError, ValueError, KeyError):
                old = {}
            try:
                new = json.loads(body)
                is_json = True
            except (TypeError, ValueError, KeyError):
                new = body
                is_json = False
            new = self.on_aw.put_properties(path=path, old=old, new=new)
            if new is None:
                self.response.set_status(400, 'Payload is not accepted')
                return
            if is_json:
                myself.property[name] = json.dumps(new)
            else:
                myself.property[name] = new
            myself.register_diffs(target='properties', subtarget=name, blob=body)
            self.response.set_status(204)
            return
        # Keep text blob for later diff registration
        blob = body
        # Make store var to be merged with original struct
        try:
            body = json.loads(body)
        except (TypeError, ValueError, KeyError):
            pass
        store = {path[len(path) - 1]: body}
        # logging.debug('store with body:' + json.dumps(store))
        # Make store to be at same level as orig value
        i = len(path)-2
        while i > 0:
            c = copy.copy(store)
            store = {path[i]: c}
            # logging.debug('store with i=' + str(i) + ' (' + json.dumps(store) + ')')
            i -= 1
        # logging.debug('Snippet to store(' + json.dumps(store) + ')')
        orig = myself.property[name]
        logging.debug('Original value(' + orig + ')')
        try:
            orig = json.loads(orig)
            merge_dict(orig, store)
            res = orig
        except (TypeError, ValueError, KeyError):
            res = store
        res = self.on_aw.put_properties(path=path, old=orig, new=res)
        if res is None:
            self.response.set_status(400, 'Payload is not accepted')
            return
        res = json.dumps(res)
        logging.debug('Result to store( ' + res + ') in /properties/' + name)
        myself.property[name] = res
        myself.register_diffs(target='properties', subtarget=name, resource=resource, blob=blob)
        self.response.set_status(204)

    def post(self, actor_id, name):
        (myself, check) = auth.init_actingweb(appreq=self,
                                              actor_id=actor_id, path='properties', subpath=name,
                                              config=self.config)
        if not myself or check.response["code"] != 200:
            return
        if not check.check_authorisation(path='properties', subpath=name, method='POST'):
            self.response.set_status(403)
            return
        if len(name) > 0:
            self.response.set_status(400)
        pair = dict()
        # Handle the simple form
        if self.request.get("property") and self.request.get("value"):
            val = self.on_aw.post_properties(prop=self.request.get("property"), data=self.request.get("value"))
            if val is None:
                self.response.set_status(403)
                return
            pair[self.request.get("property")] = val
            myself.property[self.request.get("property")] = self.request.get("value")
        elif len(self.request.arguments()) > 0:
            for name in self.request.arguments():
                val = self.on_aw.post_properties(prop=name, data=self.request.get(name))
                if val is None:
                    continue
                pair[name] = val
                myself.property[name] = val
        else:
            try:
                params = json.loads(self.request.body.decode('utf-8', 'ignore'))
            except (TypeError, ValueError, KeyError):
                self.response.set_status(400, "Error in json body")
                return
            for key in params:
                val = self.on_aw.post_properties(prop=key, data=params[key])
                if val is None:
                    continue
                pair[key] = val
                if isinstance(val, dict):
                    text = json.dumps(val)
                else:
                    text = val
                myself.property[key] = text
        if not pair:
            self.response.set_status(403, "No attributes accepted")
            return
        out = json.dumps(pair)
        myself.register_diffs(target='properties', blob=out)
        self.response.write(out.encode('utf-8'))
        self.response.headers["Content-Type"] = "application/json"
        self.response.set_status(201, 'Created')

    def delete(self, actor_id, name):
        (myself, check) = auth.init_actingweb(appreq=self,
                                              actor_id=actor_id, path='properties', subpath=name,
                                              config=self.config)
        if not myself or check.response["code"] != 200:
            return
        resource = None
        if not name:
            path = []
        else:
            path = name.split('/')
            name = path[0]
            if len(path) >= 2 and len(path[1]) > 0:
                resource = path[1]
        if not check.check_authorisation(path='properties', subpath=name, method='DELETE'):
            self.response.set_status(403)
            return
        if not name:
            if self.on_aw.delete_properties(path=path, old=myself.get_properties(), new=dict()) is False:
                self.response.set_status(403)
                return
            myself.delete_properties()
            myself.register_diffs(target='properties', subtarget=None, blob='')
            self.response.set_status(204)
            return
        if len(path) == 1:
            if self.on_aw.delete_properties(path=path, old=myself.property[name], new=dict()) is False:
                self.response.set_status(403)
                return
            myself.property[name] = None
            myself.register_diffs(target='properties', subtarget=name, blob='')
            self.response.set_status(204)
            return
        orig = myself.property[name]
        old = orig
        logging.debug('DELETE /properties original value(' + orig + ')')
        try:
            orig = json.loads(orig)
        except (TypeError, ValueError, KeyError):
            # Since /properties/something was handled above
            # orig must be json loadable
            self.response.set_status(404)
            return
        if not delete_dict(orig, path[1:]):
            self.response.set_status(404)
            return
        if self.on_aw.delete_properties(path=path, old=old, new=orig) is False:
            self.response.set_status(403)
            return
        res = json.dumps(orig)
        logging.debug('Result to store( ' + res + ') in /properties/' + name)
        myself.property[name] = res
        myself.register_diffs(target='properties', subtarget=name, resource=resource, blob='')
        self.response.set_status(204)



from builtins import str
from actingweb.handlers import base_handler


class CallbackOauthHandler(base_handler.BaseHandler):

    def get(self):
        if not self.request.get('code'):
            self.response.set_status(400, "Bad request. No code.")
            return
        code = self.request.get('code')
        actor_id = self.request.get('state')
        scope = self.request.get('scope')
        url = self.config.root + str(actor_id) + '/oauth?code=' + str(code)
        if scope:
            url += '&scope=' + scope
        self.response.set_redirect(url)



import logging

from actingweb import auth
from actingweb.handlers import base_handler


class OauthHandler(base_handler.BaseHandler):

    def get(self, actor_id, path):
        (myself, check) = auth.init_actingweb(appreq=self,
                                              actor_id=actor_id,
                                              path='oauth',
                                              subpath=path,
                                              config=self.config)
        if not myself or not check:
            return
        if not check.check_authorisation(path='oauth', subpath=path, method='GET'):
            self.response.set_status(403)
            return
        if check.type != 'oauth':
            self.response.set_status(403, "OAuth not enabled")
            return
        # Handle callback from oauth granter
        if self.request.get('code'):
            if not check.process_oauth_callback(self.request.get('code')):
                self.response.set_status(502, "OAuth Token Request Failed")
                return
            else:
                # Even if oauth is successful, we need to validate that the identity that did the oauth is identical
                # to the original identity that was bound to this actor.
                # The check_on_oauth_success() function returns False if identity (or
                # anything else) is wrong.
                if not self.on_aw.check_on_oauth_success(token=check.token):
                    logging.info('Forbidden identity.')
                    self.response.set_status(403, "Forbidden to this identity")
                    return

        redirect_uri = check.validate_oauth_token()
        if len(redirect_uri) > 0:
            self.response.set_redirect(redirect_uri)
            return
        if len(redirect_uri) == 0:
            self.on_aw.aw_init(auth=check, webobj=self)
            self.on_aw.actions_on_oauth_success()
            if check.set_cookie_on_cookie_redirect(self):
                return
            self.response.set_status(204, "OAuthorization Done")
            return
        logging.debug("OAuth token refresh failed")
        return



from builtins import object
import logging
import os
from pynamodb.models import Model
from pynamodb.attributes import UnicodeAttribute, BooleanAttribute
from pynamodb.indexes import GlobalSecondaryIndex, AllProjection

"""
    DbTrust handles all db operations for a trust
    Google datastore for google is used as a backend.
"""


class SecretIndex(GlobalSecondaryIndex):
    """
    Secondary index on trust
    """
    class Meta(object):
        index_name = 'secret-index'
        read_capacity_units = 2
        write_capacity_units = 1
        projection = AllProjection()

    secret = UnicodeAttribute(hash_key=True)


class Trust(Model):
    """ Data model for a trust relationship """
    class Meta(object):
        table_name = os.getenv('AWS_DB_PREFIX', 'demo_actingweb') + "_trusts"
        read_capacity_units = 5
        write_capacity_units = 2
        region = os.getenv('AWS_DEFAULT_REGION', 'us-west-1')
        host = os.getenv('AWS_DB_HOST', None)

    id = UnicodeAttribute(hash_key=True)
    peerid = UnicodeAttribute(range_key=True)
    baseuri = UnicodeAttribute()
    type = UnicodeAttribute()
    relationship = UnicodeAttribute()
    secret = UnicodeAttribute()
    desc = UnicodeAttribute()
    approved = BooleanAttribute()
    peer_approved = BooleanAttribute()
    verified = BooleanAttribute()
    verification_token = UnicodeAttribute()
    secret_index = SecretIndex()


class DbTrust(object):
    """
        DbTrust does all the db operations for trust objects

        The  actor_id must always be set.
    """

    def get(self,  actor_id=None, peerid=None, token=None):
        """ Retrieves the trust from the database

            Either peerid or token must be set.
            If peerid is set, token will be ignored.
        """
        if not actor_id:
            return None
        try:
            if not self.handle and peerid:
                logging.debug('    Retrieving trust from db based on peerid(' + peerid + ')')
                self.handle = Trust.get(actor_id, peerid, consistent_read=True)
            elif not self.handle and token:
                logging.debug('    Retrieving trust from db based on token(' + token + ')')
                res = Trust.secret_index.query(token)
                for h in res:
                    if actor_id == h.id:
                        self.handle = h
                        break
        except Trust.DoesNotExist:
            return None
        if not self.handle:
            return None
        t = self.handle
        return {
            "id": t.id,
            "peerid": t.peerid,
            "baseuri": t.baseuri,
            "type": t.type,
            "relationship": t.relationship,
            "secret": t.secret,
            "desc": t.desc,
            "approved": t.approved,
            "peer_approved": t.peer_approved,
            "verified": t.verified,
            "verification_token": t.verification_token,
        }

    def modify(self, baseuri=None,
               secret=None,
               desc=None,
               approved=None,
               verified=None,
               verification_token=None,
               peer_approved=None):
        """ Modify a trust 
        
            If bools are none, they will not be changed.
        """
        if not self.handle:
            logging.debug("Attempted modification of DbTrust without db handle")
            return False
        if baseuri and len(baseuri) > 0:
            self.handle.baseuri = baseuri
        if secret and len(secret) > 0:
            self.handle.secret = secret
        if desc and len(desc) > 0:
            self.handle.desc = desc
        if approved is not None:
            self.handle.approved = approved
        if verified is not None:
            self.handle.verified = verified
        if verification_token and len(verification_token) > 0:
            self.handle.verification_token = verification_token
        if peer_approved is not None:
            self.handle.peer_approved = peer_approved
        self.handle.save()
        return True

    def create(self, actor_id=None, peerid=None,
               baseuri='', peer_type='', relationship='',
               secret='', approved='', verified=False,
               peer_approved=False, verification_token='',
               desc=''):
        """ Create a new trust """
        if not actor_id or not peerid:
            return False
        self.handle = Trust(id=actor_id,
                            peerid=peerid,
                            baseuri=baseuri,
                            type=peer_type,
                            relationship=relationship,
                            secret=secret,
                            approved=approved,
                            verified=verified,
                            peer_approved=peer_approved,
                            verification_token=verification_token,
                            desc=desc)
        self.handle.save()
        return True

    def delete(self):
        """ Deletes the property in the database """
        if not self.handle:
            return False
        self.handle.delete()
        self.handle = None
        return True

    @staticmethod
    def is_token_in_db(actor_id=None, token=None):
        """ Returns True if token is found in db """
        if not actor_id or len(actor_id) == 0:
            return False
        if not token or len(token) == 0:
            return False
        for r in Trust.secret_index.query(token):
            if r.id != actor_id:
                continue
            else:
                return True
        return False

    def __init__(self):
        self.handle = None
        if not Trust.exists():
            Trust.create_table(wait=True)


class DbTrustList(object):
    """
        DbTrustList does all the db operations for list of trust objects

        The  actor_id must always be set. 
    """

    def fetch(self,  actor_id):
        """ Retrieves the trusts of an actor_id from the database as an array"""
        if not actor_id:
            return None
        self.actor_id = actor_id
        self.handle = Trust.scan(Trust.id == self.actor_id, consistent_read=True)
        self.trusts = []
        if self.handle:
            for t in self.handle:
                self.trusts.append({
                    "id": t.id,
                    "peerid": t.peerid,
                    "baseuri": t.baseuri,
                    "type": t.type,
                    "relationship": t.relationship,
                    "secret": t.secret,
                    "desc": t.desc,
                    "approved": t.approved,
                    "peer_approved": t.peer_approved,
                    "verified": t.verified,
                    "verification_token": t.verification_token,
                })
            return self.trusts
        else:
            return []

    def delete(self):
        """ Deletes all the properties in the database """
        self.handle = Trust.scan(Trust.id == self.actor_id, consistent_read=True)
        if not self.handle:
            return False
        for p in self.handle:
            p.delete()
        self.handle = None
        return True

    def __init__(self):
        self.handle = None
        self.actor_id = None
        self.trusts = []
        if not Trust.exists():
            Trust.create_table(wait=True)



from builtins import str
import json
import logging

from actingweb import trust
from actingweb import auth
from actingweb.handlers import base_handler


# /trust aw_handlers
#
# GET /trust with query parameters (relationship, type, and peerid) to retrieve trust relationships (auth: only creator
# and admins allowed)
# POST /trust with json body to initiate a trust relationship between this
#   actor and another (reciprocal relationship) (auth: only creator and admins allowed)
# POST /trust/{relationship} with json body to create new trust
#   relationship (see config.py for default relationship and auto-accept, no
#   auth required)
# GET /trust/{relationship}}/{actorid} to get details on a specific relationship (auth: creator, admin, or peer secret)
# POST /trust/{relationship}}/{actorid} to send information to a peer about changes in the relationship
# PUT /trust/{relationship}}/{actorid} with a json body to change details on a relationship (baseuri, secret, desc)
# (auth: creator,
#   admin, or peer secret)
# DELETE /trust/{relationship}}/{actorid} to delete a relationship (with
#   ?peer=true if the delete is from the peer) (auth: creator, admin, or
#   peer secret)

# Handling requests to trust/
class TrustHandler(base_handler.BaseHandler):

    def get(self, actor_id):
        if self.request.get('_method') == 'POST':
            self.post(actor_id)
            return
        (myself, check) = auth.init_actingweb(
            appreq=self,
            actor_id=actor_id, path='trust',
            config=self.config)
        if not myself or check.response["code"] != 200:
            return
        if not check.check_authorisation(path='trust', method='GET'):
            self.response.set_status(403)
            return
        relationship = self.request.get('relationship')
        peer_type = self.request.get('type')
        peerid = self.request.get('peerid', )

        pairs = myself.get_trust_relationships(
            relationship=relationship, peerid=peerid, trust_type=peer_type)
        if not pairs or len(pairs) == 0:
            self.response.set_status(404, 'Not found')
            return
        out = json.dumps(pairs)
        self.response.write(out)
        self.response.headers["Content-Type"] = "application/json"
        self.response.set_status(200, 'Ok')

    def post(self, actor_id):
        (myself, check) = auth.init_actingweb(
            appreq=self,
            actor_id=actor_id, path='trust',
            config=self.config)
        if not myself or check.response["code"] != 200:
            return
        if not check.check_authorisation(path='trust', method='POST'):
            self.response.set_status(403)
            return
        desc = ''
        relationship = self.config.default_relationship
        peer_type = ''
        try:
            params = json.loads(self.request.body.decode('utf-8', 'ignore'))
            if 'url' in params:
                url = params['url']
            else:
                url = ''
            if 'relationship' in params:
                relationship = params['relationship']
            if 'type' in params:
                peer_type = params['type']
            if 'desc' in params:
                desc = params['desc']
        except ValueError:
            url = self.request.get('url')
            relationship = self.request.get('relationship')
            peer_type = self.request.get('type')
        if len(url) == 0:
            self.response.set_status(400, 'Missing peer URL')
            return
        secret = self.config.new_token()
        new_trust = myself.create_reciprocal_trust(
            url=url, secret=secret, desc=desc, relationship=relationship,
            trust_type=peer_type)
        if not new_trust:
            self.response.set_status(408, 'Unable to create trust relationship')
            return
        self.response.headers["Location"] = str(self.config.root +
                                                myself.id + '/trust/' +
                                                new_trust["relationship"] +
                                                '/' + new_trust["peerid"])
        out = json.dumps(new_trust)
        self.response.write(out)
        self.response.headers["Content-Type"] = "application/json"
        self.response.set_status(201, 'Created')


# Handling requests to /trust/*, e.g. /trust/friend
class TrustRelationshipHandler(base_handler.BaseHandler):

    def get(self, actor_id, relationship):
        if self.request.get('_method') == 'POST':
            self.post(actor_id, relationship)
            return
        self.response.set_status(404, "Not found")

    def put(self, actor_id, relationship):
        (myself, check) = auth.init_actingweb(
            appreq=self,
            actor_id=actor_id,
            path='trust',
            subpath=relationship,
            add_response=False,
            config=self.config)
        if not myself:
            return
        if relationship != 'trustee':
            self.response.set_status(404, "Not found")
            return
        # Access is the same as /trust
        if not check.check_authorisation(path='trust', method='POST'):
            self.response.set_status(403)
            return
        try:
            params = json.loads(self.request.body.decode('utf-8', 'ignore'))
            if 'trustee_root' in params:
                trustee_root = params['trustee_root']
            else:
                trustee_root = ''
            if 'creator' in params:
                creator = params['creator']
            else:
                creator = None
        except ValueError:
            self.response.set_status(400, 'No json content')
            return
        if len(trustee_root) > 0:
            myself.store.trustee_root = trustee_root
        if creator:
            myself.modify(creator=creator)
        self.response.set_status(204, 'No content')

    def delete(self, actor_id, relationship):
        (myself, check) = auth.init_actingweb(
            appreq=self,
            actor_id=actor_id,
            path='trust',
            subpath=relationship,
            add_response=False,
            config=self.config)
        if not myself:
            return
        if relationship != 'trustee':
            self.response.set_status(404, "Not found")
            return
        # Access is the same as /trust
        if not check.check_authorisation(path='trust', method='DELETE'):
            self.response.set_status(403)
            return
        myself.store.trustee_root = None
        self.response.set_status(204, 'No content')

    def post(self, actor_id, relationship):
        (myself, check) = auth.init_actingweb(
            appreq=self,
            actor_id=actor_id,
            path='trust',
            subpath=relationship,
            add_response=False,
            config=self.config)
        if not myself:
            self.response.set_status(404)
            logging.debug("Got trust creation request for unknown Actor(" + str(id) + ")")
            return
        if not check.check_authorisation(path='trust', subpath='<type>', method='POST'):
            self.response.set_status(403)
            return
        try:
            params = json.loads(self.request.body.decode('utf-8', 'ignore'))
            if 'baseuri' in params:
                baseuri = params['baseuri']
            else:
                baseuri = ''
            if 'id' in params:
                peerid = params['id']
            else:
                peerid = ''
            if 'type' in params:
                peer_type = params['type']
            else:
                peer_type = ''
            if 'secret' in params:
                secret = params['secret']
            else:
                secret = ''
            if 'desc' in params:
                desc = params['desc']
            else:
                desc = ''
            if 'verify' in params:
                verification_token = params['verify']
            else:
                verification_token = None
        except ValueError:
            self.response.set_status(400, 'No json content')
            return

        if len(baseuri) == 0 or len(peerid) == 0 or len(peer_type) == 0:
            self.response.set_status(400, 'Missing mandatory attributes')
            return
        if self.config.auto_accept_default_relationship and self.config.default_relationship == relationship:
            approved = True
        else:
            approved = False
        # Since we received a request for a relationship, assume that peer has approved
        new_trust = myself.create_verified_trust(
            baseuri=baseuri, peerid=peerid, approved=approved, secret=secret,
            verification_token=verification_token, trust_type=peer_type, peer_approved=True,
            relationship=relationship, desc=desc)
        if not new_trust:
            self.response.set_status(403, 'Forbidden')
            return
        self.response.headers["Location"] = str(self.config.root + myself.id +
                                                '/trust/' + new_trust["relationship"] +
                                                "/" + new_trust["peerid"])
        out = json.dumps(new_trust)
        self.response.write(out)
        self.response.headers["Content-Type"] = "application/json"
        if approved:
            self.response.set_status(201, 'Created')
        else:
            self.response.set_status(202, 'Accepted')


# Handling requests to specific relationships, e.g. /trust/friend/12f2ae53bd
class TrustPeerHandler(base_handler.BaseHandler):

    def get(self, actor_id, relationship, peerid):
        if self.request.get('_method') == 'PUT':
            self.put(actor_id, relationship, peerid)
            return
        if self.request.get('_method') == 'DELETE':
            self.delete(actor_id, relationship, peerid)
            return
        logging.debug('GET trust headers: ' + str(self.request.headers))
        (myself, check) = auth.init_actingweb(
            appreq=self,
            actor_id=actor_id,
            path='trust',
            subpath=relationship,
            config=self.config)
        if not myself or check.response["code"] != 200:
            return
        if not check.check_authorisation(path='trust', subpath='<type>/<id>', method='GET', peerid=peerid):
            self.response.set_status(403)
            return
        relationships = myself.get_trust_relationships(
            relationship=relationship, peerid=peerid)
        if not relationships or len(relationships) == 0:
            self.response.set_status(404, 'Not found')
            return
        my_trust = relationships[0]
        # If the peer did a GET to verify
        if check.trust and check.trust["peerid"] == peerid and not my_trust["verified"]:
            dbtrust = trust.Trust(actor_id=actor_id, peerid=peerid, config=self.config)
            dbtrust.modify(verified=True)
            # verification_token = my_trust["verification_token"]
        out = json.dumps(my_trust)
        self.response.write(out)
        self.response.headers["Content-Type"] = "application/json"
        if my_trust["approved"]:
            self.response.set_status(200, 'Ok')
        else:
            self.response.set_status(202, 'Accepted')

    def post(self, actor_id, relationship, peerid):
        (myself, check) = auth.init_actingweb(
            appreq=self,
            actor_id=actor_id,
            path='trust',
            subpath=relationship,
            config=self.config)
        if not myself or check.response["code"] != 200:
            return
        try:
            params = json.loads(self.request.body.decode('utf-8', 'ignore'))
            peer_approved = None
            if 'approved' in params:
                if params['approved'] and params['approved'] is True:
                    peer_approved = True
        except ValueError:
            self.response.set_status(400, 'No json content')
            return
        if peer_approved:
            # If this is a notification from a peer about approval, we cannot check if the relationship is approved!
            if not check.check_authorisation(path='trust', subpath='<type>/<id>', method='POST', peerid=peerid,
                                             approved=False):
                self.response.set_status(403)
                return
        else:
            if not check.check_authorisation(path='trust', subpath='<type>/<id>', method='POST', peerid=peerid):
                self.response.set_status(403)
                return
        if myself.modify_trust_and_notify(relationship=relationship, peerid=peerid, peer_approved=peer_approved):
            self.response.set_status(204, 'Ok')
        else:
            self.response.set_status(500, 'Not modified')

    def put(self, actor_id, relationship, peerid):
        (myself, check) = auth.init_actingweb(
            appreq=self,
            actor_id=actor_id,
            path='trust',
            subpath=relationship,
            config=self.config)
        if not myself or check.response["code"] != 200:
            return
        if not check.check_authorisation(path='trust', subpath='<type>/<id>', method='PUT', peerid=peerid):
            self.response.set_status(403)
            return
        approved = None
        try:
            params = json.loads(self.request.body.decode('utf-8', 'ignore'))
            if 'baseuri' in params:
                baseuri = params['baseuri']
            else:
                baseuri = ''
            if 'desc' in params:
                desc = params['desc']
            else:
                desc = ''
            if 'approved' in params:
                if params['approved'] is True or params['approved'].lower() == "true":
                    approved = True
        except ValueError:
            if not self.request.get('_method') or self.request.get('_method') != "PUT":
                self.response.set_status(400, 'No json content')
                return
            if self.request.get('approved') and len(self.request.get('approved')) > 0:
                if self.request.get('approved').lower() == "true":
                    approved = True
            if self.request.get('baseuri') and len(self.request.get('baseuri')) > 0:
                baseuri = self.request.get('baseuri')
            else:
                baseuri = ''
            if self.request.get('desc') and len(self.request.get('desc')) > 0:
                desc = self.request.get('desc')
            else:
                desc = ''
        if myself.modify_trust_and_notify(relationship=relationship, peerid=peerid, baseuri=baseuri, approved=approved,
                                          desc=desc):
            self.response.set_status(204, 'Ok')
        else:
            self.response.set_status(500, 'Not modified')

    def delete(self, actor_id, relationship, peerid):
        (myself, check) = auth.init_actingweb(
            appreq=self,
            actor_id=actor_id,
            path='trust',
            subpath=relationship,
            add_response=False,
            config=self.config)
        if not myself or (check.response["code"] != 200 and check.response["code"] != 401):
            auth.add_auth_response(appreq=self, auth_obj=check)
            return
        # We allow non-approved peers to delete even if we haven't approved the relationship yet
        if not check.check_authorisation(path='trust', subpath='<type>/<id>', method='DELETE', peerid=peerid,
                                         approved=False):
            self.response.set_status(403)
            return
        is_peer = False
        if check.trust and check.trust["peerid"] == peerid:
            is_peer = True
        else:
            # Use of GET param peer=true is a way of forcing no deletion of a peer
            # relationship even when requestor is not a peer (primarily for testing purposes)
            peer_get = self.request.get('peer').lower()
            if peer_get.lower() == "true":
                is_peer = True
        relationships = myself.get_trust_relationships(
            relationship=relationship, peerid=peerid)
        if not relationships or len(relationships) == 0:
            self.response.set_status(404, 'Not found')
            return
        if is_peer:
            deleted = myself.delete_reciprocal_trust(peerid=peerid, delete_peer=False)
        else:
            deleted = myself.delete_reciprocal_trust(peerid=peerid, delete_peer=True)
        if not deleted:
            self.response.set_status(502, 'Not able to delete relationship with peer.')
            return
        self.response.set_status(204, 'Ok')



__all__ = [
    "actor",
    "attribute",
    "oauth",
    "auth",
    "aw_proxy",
    "peertrustee",
    "property",
    "subscription",
    "trust",
    "config",
    "on_aw",
    "aw_web_request"
]



from builtins import object
from actingweb import aw_web_request, on_aw as on_aw_class, config as config_class


class BaseHandler(object):

    def __init__(self,
                 webobj=aw_web_request.AWWebObj(),
                 config=config_class.Config(),
                 on_aw=on_aw_class.OnAWBase()
                 ):
        self.request = webobj.request
        self.response = webobj.response
        self.config = config
        self.on_aw = on_aw



from builtins import object
import logging


class Trust(object):

    def get(self):
        """ Retrieve a trust relationship with either peerid or token """
        if self.trust and len(self.trust) > 0:
            return self.trust
        if not self.peerid and self.token:
            self.trust = self.handle.get(actor_id=self.actor_id,
                                         token=self.token)
        elif self.peerid and not self.token:
            self.trust = self.handle.get(actor_id=self.actor_id,
                                         peerid=self.peerid)
        else:
            self.trust = self.handle.get(actor_id=self.actor_id,
                                         peerid=self.peerid,
                                         token=self.token)
        return self.trust

    def delete(self):
        """ Delete the trust relationship """
        if not self.handle:
            return False
        self.trust = {}
        return self.handle.delete()

    def modify(self, baseuri=None, secret=None, desc=None, approved=None,
               verified=None, verification_token=None, peer_approved=None):
        if not self.handle:
            logging.debug("Attempted modifcation of trust without handle")
            return False
        if baseuri:
            self.trust["baseuri"] = baseuri
        if secret:
            self.trust["secret"] = secret
        if desc:
            self.trust["desc"] = desc
        if approved is not None:
            self.trust["approved"] = approved
        if verified is not None:
            self.trust["verified"] = verified
        if verification_token:
            self.trust["verification_token"] = verification_token
        if peer_approved is not None:
            self.trust["peer_approved"] = peer_approved
        return self.handle.modify(baseuri=baseuri, secret=secret, desc=desc, approved=approved,
                                  verified=verified, verification_token=verification_token,
                                  peer_approved=peer_approved)

    def create(self, baseuri='', peer_type='', relationship='', secret='',
               approved=False, verified=False, verification_token='',
               desc='', peer_approved=False):
        """ Create a new trust relationship """
        self.trust = {"baseuri": baseuri, "type": peer_type}
        if not relationship or len(relationship) == 0:
            self.trust["relationship"] = self.config.default_relationship
        else:
            self.trust["relationship"] = relationship
        if not secret or len(secret) == 0:
            self.trust["secret"] = self.config.new_token()
        else:
            self.trust["secret"] = secret
        # Be absolutely sure that the secret is not already used
        testhandle = self.config.DbTrust.DbTrust()
        if testhandle.is_token_in_db(actor_id=self.actor_id, token=self.trust["secret"]):
            logging.warning("Found a non-unique token where it should be unique")
            return False
        self.trust["approved"] = approved
        self.trust["peer_approved"] = peer_approved
        self.trust["verified"] = verified
        if not verification_token or len(verification_token) == 0:
            self.trust["verification_token"] = self.config.new_token()
        self.trust["desc"] = desc
        self.trust["id"] = self.actor_id
        self.trust["peerid"] = self.peerid
        return self.handle.create(actor_id=self.actor_id,
                                  peerid=self.peerid,
                                  baseuri=self.trust["baseuri"],
                                  peer_type=self.trust["type"],
                                  relationship=self.trust["relationship"],
                                  secret=self.trust["secret"],
                                  approved=self.trust["approved"],
                                  verified=self.trust["verified"],
                                  peer_approved=self.trust["peer_approved"],
                                  verification_token=self.trust["verification_token"],
                                  desc=self.trust["desc"])

    def __init__(self, actor_id=None, peerid=None, token=None, config=None):
        self.config = config
        self.handle = self.config.DbTrust.DbTrust()
        self.trust = {}
        if not actor_id or len(actor_id) == 0:
            logging.debug("No actorid set in initialisation of trust")
            return
        if not peerid and not token:
            logging.debug("Both peerid and token are not set in initialisation of trust. One must be set.")
            return
        if not token and (not peerid or len(peerid) == 0):
            logging.debug("No peerid set in initialisation of trust")
            return
        self.actor_id = actor_id
        self.peerid = peerid
        self.token = token
        self.get()


class Trusts(object):
    """ Handles all trusts of a specific actor_id

        Access the indvidual trusts in .dbtrusts and the trust data
        in .trusts as a dictionary
    """

    def fetch(self):
        if self.trusts is not None:
            return self.trusts
        if not self.list:
            self.config.DbTrust.DbTrustList()
        if not self.trusts:
            self.trusts = self.list.fetch(actor_id=self.actor_id)
        return self.trusts

    def delete(self):
        if not self.list:
            logging.debug("Already deleted list in trusts")
            return False
        self.list.delete()
        return True

    def __init__(self,  actor_id=None, config=None):
        """ Properties must always be initialised with an actor_id """
        self.config = config
        if not actor_id:
            self.list = None
            logging.debug("No actor_id in initialisation of trusts")
            return
        self.list = self.config.DbTrust.DbTrustList()
        self.actor_id = actor_id
        self.trusts = None
        self.fetch()



from setuptools import setup


def readme():
    with open('README.rst') as f:
        return f.read()


setup(
    name='actingweb',
    version='2.6.0',
    description='The official ActingWeb library',
    long_description=readme(),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Topic :: System :: Distributed Computing',
    ],
    url='http://actingweb.org',
    author='Greger Wedel',
    author_email='support@greger.io',
    license='BSD',
    packages=[
        'actingweb',
        'actingweb.handlers',
        'actingweb.db_dynamodb'
    ],
    python_requires='>=2.7',
    install_requires=[
        'pynamodb',
        'boto3',
        'urlfetch'
    ],
    include_package_data=True,
    zip_safe=False)



from actingweb import auth
from actingweb.handlers import base_handler


class BotHandler(base_handler.BaseHandler):

    def post(self, path):
        """Handles POST callbacks for bots."""

        if not self.config.bot['token'] or len(self.config.bot['token']) == 0:
            self.response.set_status(404)
            return
        check = auth.Auth(actor_id=None, config=self.config)
        check.oauth.token = self.config.bot['token']
        # Here we need to manually initialise on_aw as this is normally done in init_actingweb() in auth
        self.on_aw.aw_init(auth=check, webobj=self)
        ret = self.on_aw.bot_post(path=path)
        if ret and 100 <= ret < 999:
            self.response.set_status(ret)
            return
        elif ret:
            self.response.set_status(204)
            return
        else:
            self.response.set_status(404)
            return
